
/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <cassert>

#include "state.h"
#include "plan.h"
#include "task.h"
#include "utils.h"
#include "fd/fd_state.h"
#include "fd/fd_globals.h"
#include "fd/fd_domain_transition_graph.h"
#include "fd/causal_graph.h"
#include "fd/cg_heuristic.h"
#include "fd/ff_heuristic.h"
#include "fd/heuristic.h"

using namespace std;

static Task&task = Task::getInstance();

static FAST_DOWNWARD::Heuristic *cgh = new FAST_DOWNWARD::CGHeuristic; 
static FAST_DOWNWARD::Heuristic *ffh = new FAST_DOWNWARD::FFHeuristic;

int get_cg_heuristic(State& orig_state){

	assert( orig_state.gvalue!= NULL);
	
	vector<int> assignment;
	for( unsigned i=0;i<task.transGraphNum;i++ )
		assignment.push_back( orig_state.gvalue[i] );
	
	FAST_DOWNWARD::State fd_state( assignment );
	
	int value = cgh->evaluate( fd_state );
	//cout<<"Downward Heuristic:"<< value << endl;
	//if( value == -1 )
	//	value = INF;
	
	return value;
}



int get_ff_heuristic(State& orig_state){

	vector<int> assignment;
	for( unsigned i=0;i<task.transGraphNum;i++ )
		assignment.push_back( orig_state.gvalue[i] );
	
	FAST_DOWNWARD::State fd_state( assignment );
	
	int value = ffh->evaluate( fd_state );
	//if( value == -1 )
	//	value = INF;
	
	return value;
}


void read_causal_graphs(bool** matrix, unsigned* mapping, int var_count){
	
	//assert( (unsigned)var_count == FAST_DOWNWARD::g_variable_domain.size() );
	
	
  	for(int from_node = 0; from_node < (int)FAST_DOWNWARD::g_variable_domain.size(); from_node++) {
  		vector<int> vec = FAST_DOWNWARD::g_causal_graph->get_successors( from_node );
  		for(unsigned i=0;i<vec.size();i++){
  			int to_node = vec[i];
  			
  			//!! The dependency in DPlan is in reversed order with downward;
  			matrix[to_node][from_node] = true;
  		}	
  	}
}
