/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/




#include<iostream>
#include<vector>
#include<map>
#include<set>

#include "dp.h"
#include "state.h"
#include "plan.h"
#include "task.h"
#include "utils.h"
#include "search_entities.h"
#include "search_context.h"
#include "fringe.h"

using namespace std;

static Task&task = Task::getInstance();


/**************************************************************************************/

FringeBucket::FringeBucket(){
	heuristic = UNDEFINED;
}


bool FringeBucket::push(FringeElem& fe){
	

	list<FringeElem>::iterator iter = elems.begin();
	for(;iter!=elems.end();iter++){
		if( fe.edge->graph->type_identifier ==  iter->edge->graph->type_identifier){
			return false;
		}
	} 
	
	elems.push_back( fe );
	return true;
}


unsigned FringeBucket::shrink(){
	
	unsigned count = (int)elems.size();
	count /= 2;
	for(unsigned i=0; i<count;i++ )
		elems.pop_back();
	return count; 
	
}

void FringeBucket::print(){

	list<FringeElem>::iterator iter = elems.begin();
	cout<<"{B"<<heuristic<<":";
	for( ; iter!=elems.end();iter++){
		printf("[%d/%d:%u]", iter->heuristic, iter->plan.cost(), 
				iter->hash );
	}
	cout<<"}"<<endl;
}


/**************************************************************************************/


FringeSlot::FringeSlot(){
	cost = UNDEFINED;
	elem_count = 0;
}


unsigned FringeSlot::count(){
	return elem_count;
}


bool FringeSlot::find_bucket(unsigned h_value, int& pos, bool debug){
	assert( h_value != UNDEFINED );
	if( buckets.size() == 0 ){
		pos =0;
		return false;
	}
	
    int low = 0;
    int high = (int)this->buckets.size()-1;
    while (low < high) {
        int mid = (low + high)/2;
        if ( buckets[mid].heuristic < h_value )
            low = mid + 1;
        else
        	high = mid;
    }
 
    pos = low;
    assert( buckets.size()==0 || low < (int)buckets.size() );
    assert( low >= (int)buckets.size() || buckets[low].heuristic != UNDEFINED );
    
    //Exception of pos after binary search;
    if( h_value > buckets[buckets.size()-1].heuristic ) pos =  buckets.size();
    
    if ( low < (int)buckets.size() &&  buckets[low].heuristic == h_value )
    	return true;
    else
    	return false;
}



bool FringeSlot::push(FringeElem& fe){

	unsigned h_value = fe.heuristic;
	assert( h_value != UNDEFINED );
	
	int pos;
    bool found = find_bucket( h_value, pos );

    assert( pos >=0 && ( pos ==0 || pos<=(int)buckets.size()));
    
    bool succ = false;
    if ( found )
    	succ = buckets[pos].push( fe );
    else{
    	//When an appropriate slot is not found;
    	FringeBucket h_bucket;
    	h_bucket.heuristic = h_value;
    	succ = h_bucket.push( fe );
    	buckets.insert( buckets.begin() + pos , h_bucket );
    }
    
   	if( succ )
   		elem_count++;
   	return succ;
}


void FringeSlot::print(){
	cout<<endl<<endl<<endl;
	for(unsigned j=0;j<buckets.size();j++){
		list<FringeElem>::iterator iter = buckets[j].elems.begin();
		printf("{%d}", buckets[j].heuristic );
		for(;iter!=buckets[j].elems.end();iter++){ 
			printf("[%d/%d:%u]", iter->heuristic , 
				iter->plan.cost(), iter->hash );
		}
		cout<<";";
	}
	cout<<endl;	
}



/*********************************************
 * 				Close List			 		**
 * *******************************************/

/*
 * Returns true or false, depends on existing or not;
 */
bool CloseList::add( State& state ){
	
	assert( state.gvalue!= NULL );
	unsigned hv = state.hash_value();
	
	if( data.size() == 0 ){
		data.insert( data.begin() , hv );
		return false;
	}
	
	set<unsigned>::iterator iter = data.find( hv );
	if( iter == data.end() ){
		data.insert( hv );
		return false;
	}else
		return true;
}

bool CloseList::exists ( State& state ){
	
	if( state.gvalue == NULL ) return false;
	if( state.gvalue[0] == UNDEFINED ) return false;
	unsigned hv = state.hash_value();

	
	if( data.size() == 0 )
		return false;

	set<unsigned>::iterator iter = data.find( hv );
	if( iter == data.end() )
		return false;
	else
		return true;
	
}

void CloseList::clear(){
	data.clear();
}

void CloseList::print_info(){
	cout<<"Size of the close list:["<<data.size()<<"]."<<endl;
}

//END of file;


