#ifndef FD_OPERATOR_H
#define FD_OPERATOR_H

#include <cassert>
#include <iostream>
#include <string>
#include <vector>

#include "fd_globals.h"
#include "fd_state.h"

namespace FAST_DOWNWARD{

class Variable;

struct Prevail {
    int var;
    int prev;
    Prevail(std::istream &in);
    Prevail(int v, int p) : var(v), prev(p) {}

    bool is_applicable(const State &state) const {
	assert(var >= 0 && var < (int)g_variable_name.size());
	assert(prev >= 0 && prev < (int)g_variable_domain[var]);
	return state[var] == prev;
    }

    void dump() const;
};

struct PrePost {
    int var;
    int pre, post;
    std::vector<Prevail> cond;
    PrePost() {} // Needed for axiom file-reading constructor, unfortunately.
    PrePost(std::istream &in);
    PrePost(int v, int pr, int po, const std::vector<Prevail> &co)
	: var(v), pre(pr), post(po), cond(co) {}

    bool is_applicable(const State &state) const {
		assert(var >= 0 && var < (int)g_variable_name.size());
		assert(pre == -1 || pre >= 0 && pre < (int)g_variable_domain[var]);
		return pre == -1 || state[var] == pre;
    }

    bool does_fire(const State &state) const {
	for(int i = 0; i < (int)cond.size(); i++)
	    if(!cond[i].is_applicable(state))
		return false;
	return true;
    }

    void dump() const;
};

class Operator {
    bool is_an_axiom;
    std::vector<Prevail> prevail;      // var, val
    std::vector<PrePost> pre_post;     // var, old-val, new-val, effect conditions
    std::string name;
    unsigned cost;    
public:	
	unsigned get_cost(){ return cost;};
    Operator(std::istream &in, bool is_axiom);
    void dump() const;
    std::string get_name() const {return name;}

    bool is_axiom() const {return is_an_axiom;}

    const std::vector<Prevail> &get_prevail() const {return prevail;}
    const std::vector<PrePost> &get_pre_post() const {return pre_post;}

    bool is_applicable(const State &state) const {
		for(int i = 0; i < (int)prevail.size(); i++)
		    if(!prevail[i].is_applicable(state))
			return false;
		for(int i = 0; i < (int)pre_post.size(); i++)
		    if(!pre_post[i].is_applicable(state))
			return false;
		return true;
    }
};


}
#endif
