#ifndef FF_HEURISTIC_H
#define FF_HEURISTIC_H

#include "heuristic.h"
#include "fd_operator.h"

#include <vector>
#include <ext/hash_set>

namespace FAST_DOWNWARD{


class Operator;
class State;

class Proposition;
class UnaryOperator;

struct UnaryOperator {
    Operator *op;
    std::vector<Proposition *> precondition;
    Proposition *effect;
    int base_cost; // 0 for axioms, 1 for regular operators
    int base_op; // 1 for regular operators, 0 for axiom
    
    int unsatisfied_preconditions;
    int h_add_cost;
    int needed_ops; // used for count ops
    UnaryOperator(const std::vector<Proposition *> &pre, Proposition *eff,
		  Operator *the_op, int base) : op(the_op), precondition(pre), effect(eff), 
		  base_cost(base) {
		if ( op->is_axiom() ) 
			base_op = 0;
		else
			base_op = 1;
	}
};

struct Proposition {
    bool is_goal;
    std::vector<UnaryOperator *> precondition_of;

    int h_add_cost;
    int needed_ops; // use for count ops
    UnaryOperator *reached_by;

    Proposition() {
			is_goal = false;
			h_add_cost = -1;
			reached_by = 0;
    }
};

struct hash_operator_ptr {
    size_t operator()(Operator *key) const {
	return reinterpret_cast<unsigned long>(key);
    }
};

class FFHeuristic : public Heuristic {
    typedef __gnu_cxx::hash_set<Operator *, hash_operator_ptr> RelaxedPlan;

    std::vector<UnaryOperator> unary_operators;
    std::vector<std::vector<Proposition> > propositions;
    std::vector<Proposition *> goal_propositions;

    Proposition **reachable_queue_start;
    Proposition **reachable_queue_read_pos;
    Proposition **reachable_queue_write_pos;

    void build_unary_operators(Operator &op);
    void simplify();

    void setup_exploration_queue();
    void setup_exploration_queue_state(const State &state);
    void relaxed_exploration();
    void collect_relaxed_plan(Proposition *goal, RelaxedPlan &relaxed_plan);

    int compute_hsp_add_heuristic();
    int compute_ff_heuristic();
protected:
    virtual void initialize();
    virtual int compute_heuristic(const State &state);
public:
    FFHeuristic(bool use_cache=false);
    ~FFHeuristic();
};



}
#endif
