#ifndef FRINGE_H
#define FRINGE_H


#include<iostream>
#include<vector>
#include<list>
#include<map>
#include<set>

#include "dp.h"

class SearchContext;
class State;
class Plan;
class FringeSlot;
class Fringe;

typedef pair<unsigned,unsigned> FRINGE_INDEX_TYPE;	//< g+ h, h>


class FringeElem{
public:
	Plan  	 plan;

	unsigned hash;
	unsigned heuristic;
	TransitionEdge* edge;
	bool privi;
	
	FringeElem(Plan& p){
		heuristic = UNDEFINED;
		hash = UNDEFINED;
		privi = false;
		edge = NULL;
		plan = p;
	}
};


class FringeBucket{
public:
	list<FringeElem> elems;
	FringeSlot*  parent;
	unsigned     heuristic;
	
	FringeBucket();
	FringeElem& operator[](unsigned offset);
	bool push(FringeElem& fe);
	unsigned shrink();
	void print();
};


class FringeSlot{
public:
	vector<FringeBucket> buckets;
	Fringe*      parent;
	unsigned     cost;
	unsigned 	 elem_count;
	
	FringeSlot();
	FringeBucket& operator[](unsigned offset);
	bool push(FringeElem& fe);
	bool find_bucket(unsigned h_value, int& pos, bool debug = false);
	void print();
	unsigned count();
};


class CloseList{
	set<unsigned> data;
		
public:
	bool add( State& state );
	bool exists ( State& state );
	void print_info();
	void clear();
};


class Fringe{
	vector<FringeSlot> slots;
	SearchContext& context;	
	vector<unsigned>& goals;
	map<unsigned, FRINGE_INDEX_TYPE > index;
	map<unsigned, unsigned > deferred_index;
	unsigned elem_count;
	vector<unsigned> selected_graphs;
public:
	
	
	Fringe(SearchContext& sc, vector<unsigned>& gvec);
	~Fringe();
	FringeElem 	pop();
		bool 	insert(FringeElem& node);	
		bool	find_slot(unsigned cost, int& pos);
	unsigned	shrink();
	unsigned 	expand_all_succ(FringeElem& f_elem, unsigned scope);
	unsigned 	compute_heuristic(FringeElem& elem);
	    bool 	constains(State& state);
		bool 	is_empty();
	unsigned	count();
		bool	select_graphs(unsigned scope);
		
	//Statistical info:
	unsigned expand_count;
	unsigned expand_conflict;
	unsigned href_count;

	void 	print_status();
	void 	check_count_consistency();
	void 	print_statistics();
};


#endif 
