#ifndef SEARCH_ENTITIES_H
#define SEARCH_ENTITIES_H

#include<iostream>
#include<vector>

const unsigned MAX_FACTSET_SIZE = 200;

using namespace std;

class Plan;
class Task;
class State;
class TransitionGraph;
class TransitionEdge;
class SearchContext;




class FactsetSearch{

public:	
	SearchContext& 	context;
	unsigned 		fromAction;
	unsigned		factset[MAX_FACTSET_SIZE];
	unsigned		size;
	unsigned		hash_point;
	
	FactsetSearch(unsigned* facts, unsigned size, SearchContext& sc, unsigned fa);
	unsigned translate_facts( unsigned *graphs,  unsigned* vertice );
	Plan solve_fixed_order(	State& pis, unsigned* gIndex, unsigned *vIndex	);
											
	Plan solve_incremental(State& passed_in_state);
	Plan solve_atomic(State& passed_in_state);
	
	Plan solve( State& passed_in_state );
};


class DTGSearch{
public:
	TransitionGraph& graph;
	unsigned fromIndex, toIndex;
	unsigned fromFact, toFact;
	unsigned hash_point;	//h.s.p.
	SearchContext& context;
	
	DTGSearch(TransitionGraph&g, SearchContext& sc)
				:graph(g),context(sc){
		toFact    = UNDEFINED;
		fromFact  = UNDEFINED;
		fromIndex = UNDEFINED;
		toIndex   = UNDEFINED;
		hash_point = 0;
	}

	void set_dest( unsigned factId);
	void set_dest_index(unsigned index );
	
	Plan solve( Plan& plan );				
	Plan solve_path_no_merge(	Plan& plan,
								unsigned **path,
								unsigned *path_len,
								unsigned path_num);
								
	Plan solve_edge(TransitionEdge& e, State& passed_in_state );	
	void process_action_list(	TransitionEdge& edge,
								State& currentState, 
								vector<unsigned>& actionList );	
};


class GoalSearch{
	Plan best_soln;
	unsigned soln_count;
		bool gt_acco;	
		bool skip_first_ss;
		
	Plan solve_one_accompany_goal(State& state, SearchContext& context, unsigned);
public:

	vector<unsigned> orig_goals;
	vector<unsigned> goals;
	
	GoalSearch(vector<unsigned> g_vec);
	void reduce_goals();
	void record_privilege_states(Plan& plan, State& state);
	
	//Operations
	Plan solve(State& currentState);
	//Plan solve_blind_search(State& state, unsigned index);		deprecated;
	Plan solve_guided_search(State& state, unsigned scope);
	Plan solve_atomic_fixed_order(State& state);
	Plan solve_atomic_accompany_goals(State& state);
	
	//Update solutions;
	bool update_solution(Plan& plan);		
};

#endif 













