#ifndef TASK_H
#define TASK_H

#include<string>
#include<vector>
#include<iostream>
#include<set>
#include<stdio.h>
#include<sys/timeb.h>

#include "dp.h"

using namespace std;

class SolverContext;
class Plan;
class TransitionEdge;
class TransitionGraph;

class Object{
public:
	string name;
	int type;		//reference to which type the object is;
	
	unsigned* F;
	unsigned num_F;
	
	Object(string name){
		this->name = name ;
		num_F = 0;
		F = NULL;
		type = UNDEFINED;
	}
	~Object(){
		if(F!=NULL) free(F);
	}
};


/*						Basic Problem defintion strcutures		*/


class Fact{
public:
	string 		name;
	unsigned 	rand;
	unsigned	predicate;
	
	unsigned num_P;
	unsigned* P;
	unsigned num_D;
	unsigned* D;
	unsigned num_A;
	unsigned* A;
	
	unsigned graph;			//initialized as $UNDEFINED
	unsigned vertex;		//initilized as $UNDEFINED
	
	void print_detailed();
	void print_name(){cout<<name;}
	Fact(string name){this->name = name;};
	Fact(){
		num_P = num_D = num_A = 0;
		P = D = A = NULL;
		graph = vertex = INVALID;
		predicate = UNDEFINED;
	};
	
	string format();
	bool has_A();
	bool has_P();
	bool has_D();
	
	~Fact(){
		free( P );
		free( A );
		free( D );
	}
};


class Operator{
public:
	string name;
	unsigned cost;
	
	unsigned num_P;
	unsigned* P;
	unsigned num_NP;
	unsigned* NP;
	unsigned num_D;
	unsigned* D;
	unsigned num_A;
	unsigned* A;
	bool has_A(int ft);
	bool has_P(int ft);
	bool has_D(int ft);
	
	unsigned num_T;
	unsigned* Tg;
	unsigned* Tp;
	unsigned* Te;
	
	bool isDisabled;
	
	Operator(){
		num_T = num_D = num_P = num_A = num_NP = 0;
		D = P = A = NP = NULL;
		Tg = Tp = Te = NULL;
		isDisabled = false;
		cost = 1;
	}
	
	~Operator(){
		free( P );
		free( NP );
		free( D ) ;
		free( A );
		free( Tg );
		free( Tp );
		free( Te );
	}
	
	void parse(ifstream &ifs);
	void print_str();
	void print_int();
	void print_name(){cout<<name;};
	void print_as_sas();
	void print_detail();
};


typedef pair< unsigned,vector<unsigned> > AccoPairs;
typedef set< pair<unsigned,unsigned> > MutexSetType;

class Task{
private:
	Task(){
		gnum_op = gnum_ft = 0;
		transGraphNum = 0;
		fact_ordering = NULL;
		gop = NULL;
		gft = NULL;
		transGraph = NULL;
		transitCausalMatrix = causalMatrix = NULL;
	};
	
	void parse();	
	void action_rep_transform();
	void build_dependencies();
	void disable_unreachable_actions();
	
	//Private, for mapping between Task and Downward's var ordering only
	unsigned var_mapping[MAX_DTG_NUM]; 
	unsigned var_mapping_count;
public:
	~Task();
	void parse_forced_ordering();
	void parse_fast_downward();	
	void parse_axioms();
	static Task& getInstance(){
		static Task instance;
		return instance;
	};
	SolverContext* solverContext;
	
	//Problem instance information;
	Operator* 	gop;
	Fact* 		gft;
	unsigned 	gnum_op;
	unsigned 	gnum_ft;
	unsigned 	gpos_xm;
	vector<string> types;
	vector<string> predicates;	
	vector<Object> objects;
	vector<unsigned> orig_goals;
	vector<unsigned> goal;
	vector<unsigned> init;
	
	//Additional Infomations read from Downward Parser;
	vector< vector<unsigned> > graphGroup;							
	vector< vector<unsigned> > mutexGroup;
	set< pair<unsigned,unsigned> > mutex_set;
	bool ** fact_ordering;
	vector<vector<unsigned> > goals_agenda;
	vector<AccoPairs> aco_pr_vec;
	
	
	//Domain transition Graphs;
	TransitionGraph* transGraph;
	unsigned 	transGraphNum;
	bool** 		causalMatrix;
	bool**		transitCausalMatrix;
	
	
	//major Routines
	void initialize_everything();//Must call after parsed in;	
	Plan solve();
	void build_sas_formulation();
	void ordering_analysis();
	void goal_agenda_analysis();
	void goal_implications();
	void accompany_relations_analysis();
	void analysis();
	

	//Aux routins
	vector<unsigned> get_enforced_preconditions(unsigned fact);	
	int find_fact(string str);	//return -1 if not found;
	bool detect_deadend( vector<unsigned> test_set);	
	bool test_reachability(State& state, vector<unsigned> fixed_facts, unsigned goal);

	
	//For printing Out;
	void print_types();
	void print_objects();
	void print_goals();
	void print_all_gfts();
	void print_all_gops();
	void print();
	
	void print_ft_name(unsigned ft);
	void print_op_name(unsigned op){cout<<this->gop[op].name;}
	void print_ft_name(unsigned g, unsigned v);
	void print_acco_relations();
};


#endif 
