#ifndef TRANSITION_GRAPH_H
#define TRANSITION_GRAPH_H

#include<cassert>
#include<vector>
#include<string>

#include "dp.h"
//#include "task.h"

using namespace std;

const unsigned MAX_FP_NUM = 10;

class TransitionGraph;
class Plan;
class Task;
class State;

class TransitionVertex{
public:
	unsigned dimension;
	unsigned* elems;
	TransitionGraph* graph;

	void set_value(vector<int>& elems);
	void set_value(int elem);
	bool has_fact(unsigned fact);
	unsigned& operator[](unsigned offset);
	void print();
	void println();
	bool isNone();

	TransitionVertex(){
		elems = new unsigned[10];
	}
	~TransitionVertex(){
		delete[] this->elems;
	}	
};

class TransitionEdge{
public:
	TransitionGraph* graph;
	int  fromIndex,toIndex;
	unsigned num_action;
	unsigned* action; 
	
	unsigned num_SP;
	unsigned SP[5];
	unsigned num_FP;
	unsigned FP[MAX_FP_NUM];	//to-do:May need change to dynamic-allocation here;
	
	void print();
	void println();
	void println(const char*);
	
	TransitionEdge(){
		action = NULL;
		num_SP = 0;
		num_FP = 0;
		num_action = 0;
	};


	~TransitionEdge(){
		if( action != NULL) delete[] action;
	}
};



class TransitionGraph{
public:
	Task* ptr_task;
	//	Basic Properties
	unsigned id;
	unsigned invariantId;
	unsigned dimension;
	unsigned size;
	string 	 actor;
	
	//Other characteristics;
	bool 	 isMono;				//Useful while detecting goal-transforming;
	bool 	 isAxiom;
	unsigned type_identifier;
	//negateAll is the "NONE-" vertex. equals to -1 means doesn't have;
	int      hasNegateAll;
	
	int		 dependToDegree;
	int 	 dependByDegree;
	unsigned dlevel;	
	
	//	Main Structures
	TransitionVertex* 	vertex;
	TransitionEdge** 	edge;
	Plan** 				path;
	unsigned** 			distance;
	
	vector<unsigned> dependencyList;
	
	void calc_distances();
	void build_vertexes(vector<unsigned>& facts);
	void build_edges();
	void build_inner_path();
	unsigned compute_identifier();
		
	// Indexing and Printing functions;
	bool has_fact(unsigned fact);
	bool has_oper(unsigned oper);
	int  depend_degree();
	
	void print();
	void print_with_full_detail();
	void print_vertex_only();
	void print_with_inner_path();
	void print_vertex(int index);
	void print_edge(int from, int to);
	
	/* DTG_searching helpers*/
	unsigned retrieve_all_path( unsigned fromIndex, unsigned toIndex, 
							unsigned** pathList, unsigned* len);	
	void dtg_path_depth_search( unsigned currentIndex,  unsigned goalIndex, 
							unsigned ** pathList, unsigned *len,  unsigned *pathNum);
	
	TransitionGraph();
	~TransitionGraph();
};

#endif 
