/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include<cassert>
#include<iostream>
#include<fstream>
#include<vector>

#include "state.h"
#include "plan.h"
#include "transition_graph.h"
#include "task.h"

using namespace std;

static Task& task = Task::getInstance();

void Operator::print_int(){
	unsigned i;
	cout<<name<<endl;
	cout<<"PRE:";
	for(i=0;i<num_P;i++)
		cout<<P[i]<<" ";
	cout<<"DEL:";
	for(i=0;i<num_D;i++)
		cout<<D[i]<<" ";
	cout<<"ADD:";
	for(i=0;i<num_A;i++)
		cout<<A[i]<<" ";
	cout<<endl;
	for(i=0;i<num_T;i++)
		cout<<"["<<Tg[i]<<" "<<Tp[i]<<" "<<Te[i]<<"]";
	cout<<endl;
}


void Operator::print_detail(){
	unsigned i;
	cout<<name<<endl;
	cout<<"PRE:";
	for(i=0;i<num_P;i++){
		task.print_ft_name( P[i] );
		cout<<";";
	}
	cout<<"Neg-PRE:";
	for(i=0;i<num_NP;i++){
		task.print_ft_name( NP[i] );
		cout<<";";
	}	
	cout<<endl<<"DEL:";
	for(i=0;i<num_D;i++){
		task.print_ft_name( D[i] );
		cout<<";";
	}
	cout<<endl<<"ADD:";
	for(i=0;i<num_A;i++){
		task.print_ft_name( A[i] );
		cout<<";";
	}
	cout<<endl;
}


void Operator::print_as_sas(){
	Task& task = Task::getInstance();
	cout<<name;
	cout<<"num_T"<<num_T<<endl;
	for(unsigned i=0;i<num_T;i++){
		cout<<"G("<<Tg[i]<<"):";
		cout<<"P("<<Tp[i]<<")";
		if( Tp[i] == NO_PRE_NEEDED )
			cout<<"NO PRE;";
		else
			task.transGraph[Tg[i]].vertex[Tp[i]].print();
		cout<<"E("<<Te[i]<<")";
		task.transGraph[Tg[i]].vertex[Te[i]].print();
		cout<<endl;
	}
}

void Operator::parse(ifstream &ifs ){

	unsigned count,j;
	int n;
	string str;
	
	getline( ifs, str);		//Read to the end of the line;
	getline( ifs, str);
	this->name = str;
	
	ifs>>cost;
	
	vector<unsigned> vec_P;
	vector<unsigned> vec_NP;
	
	//Pre-conditions
	ifs >> count;
	for(j=0;j<count;j++){
		ifs >> n;
		if( n < 0 ){ //> (int)task.gnum_ft )
			vec_NP.push_back( -1 * n - 1 );	//Hack to read-in negated pre-conditions;
		}else
			vec_P.push_back( (unsigned)n );
	}
	
	this->P = (unsigned *) malloc( vec_P.size() * (sizeof(unsigned) + 5 ) );
	this->num_P = vec_P.size();
	this->NP = (unsigned *) malloc( vec_NP.size() * sizeof(unsigned) );
	this->num_NP = vec_NP.size();
	for( j=0;j<vec_P.size() ;j++) this->P[j]  = vec_P[j];
	for( j=0;j<vec_NP.size();j++) this->NP[j] = vec_NP[j];
	
	//Add-effect
	ifs >> count;
	this->A = (unsigned *)malloc(count* sizeof(unsigned));	
	this->num_A = count;
	for(j=0;j<count;j++){
		ifs >> n;
		this->A[j] = n;
	}
		
	//Del-Effect
	ifs >> count;
	this->D = (unsigned *) malloc( count * sizeof(unsigned) );
	this->num_D = count;
	for(j=0;j<count;j++){
		ifs >> n;
		this->D[j] = n;
	}
}


