/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include<iostream>
#include<cassert>

#include "dp.h"
#include "state.h"
#include "plan.h"
#include "task.h"
#include "resource_allocator.h"

using namespace std;

const int INCREMENTAL_ALLOC_SIZE = 1000;

ResourceAllocator::ResourceAllocator(){
	pool_size = 0;
	state_unit_size = 0;
}

unsigned* ResourceAllocator::allocate_state(unsigned& pos){
	
	if( state_unit_size == 0 ){
		Task&task = Task::getInstance();
		state_unit_size = task.transGraphNum;
		assert( state_unit_size != 0 );
	}
	
	if( ready_units.size() == 0 ){
		
		for(unsigned i=pool_size;i<pool_size+INCREMENTAL_ALLOC_SIZE;i++){
			pool[i] = (unsigned* )malloc( sizeof(unsigned) * state_unit_size );
			counter[i] = 0;
			ready_units.push_back( i );
		}
		pool_size += INCREMENTAL_ALLOC_SIZE;
		assert( pool_size < MAX_POOL_SIZE );
		if( pool_size > MAX_POOL_SIZE){
			cout<<"Set a larger resource pool.(Currently "<<MAX_POOL_SIZE<<")"<<endl;
		}
	}
	
	pos = ready_units[0];
	ready_units.pop_front();
	memset( pool[pos] , UNDEFINED, sizeof(unsigned) * state_unit_size );	
	return pool[pos];
}

void ResourceAllocator::inc_state_counter(unsigned pos){
	counter[pos]++;
}

void ResourceAllocator::dec_state_counter(unsigned pos){
	
	assert( pos != UNDEFINED );
	counter[pos]--;
	assert( counter[pos] >= 0 );
	
	if( counter[pos] == 0 )
		ready_units.push_back( pos );
}

ResourceAllocator::~ResourceAllocator(){
	for(unsigned i=0;i<pool_size;i++)
		free( pool[i] );
}

//END of file;

