

#include<iostream>
#include<stdio.h>
using namespace std;


#include "dp.h"
#include "state.h"
#include "plan.h"
#include "solver_context.h"
#include "utils.h"

void print_usage(){
	
	cout<<"dplan -d domain.pddl -p problem.pddl [-i I]"<<endl;
	
	cout<<"    001 Print all task"<<endl;
	cout<<"    002 Print all types"<<endl;
	cout<<"    003 Print all objects"<<endl;
	cout<<"    004 Print all facts"<<endl;
	cout<<"    005 Print all actions"<<endl;
	cout<<"    006 Print all actions in SAS+ formulism"<<endl;
	cout<<"    007 All Mutex Groups"<<endl;
	cout<<"    008 All Graph Groups"<<endl;
	cout<<"    009 All Axioms"<<endl;
	
	//Ordering and goal preprocessing;
	cout<<"    010 Fact ordering"<<endl;
	cout<<"    011 Accompany relations"<<endl;
	cout<<"    012 Implied Goals Analysis result"<<endl;
	cout<<"    013 Goals with agenda"<<endl;
		
	//DTG part
	cout<<"    100 Print domain transition graphs"<<endl;
	cout<<"    101 Print domain transition graphs in a more verbose format"<<endl;
	cout<<"    102 Print domain transition graphs, vertexes only"<<endl;
	
	cout<<"    110 Print Dependencies of transition graphs"<<endl;
	cout<<"    111 Print Dependencies, along with the transition graphs"<<endl;
	
	//Searching part settings;
	cout<<"	   200 Print dlevels of graphs"<<endl;
	cout<<"    201 Print transformed goals"<<endl;
	
	//On-going search infos;
	cout<<"    300 Print graphs in the scope"<<endl;
}


SolverContext::SolverContext(){
	display_info = 1;
	debug = 0;
	search_print_level = 0;
	search_strategy = SS_DIRECT_ATOM;
	search_strategy = SS_HEURISTIC;
	fixed_scope = UNDEFINED;
	global_tlimit = 1800;
	dgo = true;
  	
	memset(this->domain_file, '\0', sizeof(char)* MAX_FILENAME);
	memset(this->prob_file, '\0', sizeof(char)*  MAX_FILENAME);
	memset(this->path, '\0', sizeof(char)* MAX_FILENAME);
	this->soln_file[0] =  0;
}


bool SolverContext::process_command_line(int argc, char *argv[]){
	
	//Default Settings;

	
	int i;
	for (i = 1; i < argc; i++){

		if ( argv[i][0] != '-' ) continue;
		else if( strcmp (argv[i], "--path") == 0 || strcmp (argv[i], "-path") == 0) 
			strncpy( this->path, argv[++i], MAX_FILENAME );
		else if ( strcmp (argv[i], "-d") == 0 || strcmp (argv[i], "-domain") == 0 )
			strncpy( this->domain_file, argv[++i], MAX_FILENAME );
		else if ( strcmp (argv[i], "-p") == 0 || strcmp (argv[i], "-problem") == 0 )
			strncpy( this->prob_file, argv[++i], MAX_FILENAME );
		else if ( strcmp (argv[i], "-i") == 0 )
			sscanf( argv[++i], "%d", &this->display_info );
		else if ( strcmp (argv[i], "-o") == 0 )
			strncpy( this->soln_file, argv[++i], MAX_FILENAME );
		else if ( strcmp (argv[i], "--level") == 0 )
			sscanf( argv[++i], "%d", &search_print_level );
		else if ( strcmp (argv[i], "--debug") == 0 )
			this->debug = true;
		else if ( strcmp (argv[i], "--soln") == 0 || strcmp (argv[i], "-soln") == 0)
			strncpy( this->soln_file, argv[++i], MAX_FILENAME );
		else if ( strcmp (argv[i], "-s") == 0 )
			sscanf( argv[++i], "%d", &search_strategy );
		else if ( strcmp (argv[i], "-sh") == 0 || strcmp (argv[i], "--sh") == 0)
			this->search_strategy = SS_HEURISTIC;
		else if ( strcmp (argv[i], "-sda") == 0 || strcmp (argv[i], "--sda") == 0)
			this->search_strategy = SS_DIRECT_ATOM;
		else if ( strcmp (argv[i], "-fs") == 0 || strcmp (argv[i], "--fixscp") == 0)
			sscanf( argv[++i], "%d", &fixed_scope );		
		else if ( strcmp (argv[i], "-tl") == 0 )
			sscanf( argv[++i], "%d", &global_tlimit );
		else if ( strcmp (argv[i], "-h") == 0 )
			print_usage(),exit(0);
		else if ( strcmp (argv[i], "-ndgo") == 0 || strcmp (argv[i], "--ndgo") == 0)
			this->dgo = false;
	}
	
	
	if( strlen(this->soln_file)==0 ){
		int i;
		char* filename = this->prob_file;
		for(i=strlen(filename)-1;i>=0;i--)
			if( filename[i] == '/' ) break;
		i++;
		
		char absName[1000];
		strcpy(absName,filename+i);
		
		strcpy( this->soln_file, absName);
		strcat( this->soln_file, ".soln");
	}

	if( strlen( path ) !=0 ){
		char str[MAX_FILENAME];
		memset( str, '\0', sizeof(char)*MAX_FILENAME);
		
		strncpy( str, path, strlen(path) );
		strncat( str, domain_file, strlen(domain_file));
		strncpy( domain_file, str, strlen(str));
		
		str[0] = '\0';
		strcpy( str, path);
		strncat( str, prob_file, strlen(prob_file));
		strncpy( prob_file, str, strlen(str));		
	}
	
	
	if( strlen(this->domain_file) == 0 || strlen(this->prob_file) == 0 ){
		print_usage();
		exit(0);
	}

	if( search_print_level >0 )
		debug = true;
		 	
	if( debug || search_print_level )
		setlinebuf(stdout);

	return true;
}	
	


bool SolverContext::check_mlimit(){
	
	float current = getVmRSS();
	if( current - global_mlimit > 0.001){
		printf("Max memory consumption limitation exceeded.\n");
		exit(0);
	}
	return true;
}

void SolverContext::update_memory_usage(){

// OS_CYGWIN doesn't support vmrss;
#ifndef OS_CYGWIN
	float current = getVmRSS();
	if( current - max_mcspn > 0.001)
		max_mcspn = current;
	printf( "Memory usage at this check point is %lf\n", current);	
#endif	
}

bool SolverContext::check_tlimit(){
	//long current = (long)get_global_time();
	if( get_global_time_int() > (long)global_tlimit ){
		printf("Solver failed with time out.(%u seconds limit)\n.", global_tlimit );
		exit(0);
	}
	return 0;
}



