/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include<vector>
#include<iostream>
#include<fstream>
#include<sstream>
#include<cassert>
#include<algorithm>
#include<stdio.h>
#include<stdlib.h>

#include "dp.h"
#include "state.h"
#include "plan.h"
#include "transition_graph.h"
#include "task.h"
#include "solver_context.h"
#include "search_entities.h"
#include "./fd/fd_globals.h"
#include "utils.h"

using namespace std;

Task::~Task(){

	if( gop!=NULL)
		delete[] gop;

	if( gft != NULL )
		delete[] gft;
	
	if( fact_ordering!= NULL ){
		for(unsigned i=0;i<gnum_ft;i++)
			delete[] fact_ordering[i];
		delete[] fact_ordering;
	}
	if( transGraph!=NULL )
		delete[] transGraph;
}

/*
 * The strategy can be estimated by:  
 * 	1. number of goals / number of DTG ?
 *  2. 
 */
Plan Task::solve(){

	GoalSearch goal_search(goal);
	State initState( init );

	//choose different strategies
	Plan result = goal_search.solve( initState );
	return result;
}


int Task::find_fact(string str){
	unsigned i,j;
	
	for(i=0;i<str.length();i++)
		str[i] = tolower(str[i]);

	for(i=0;i<gnum_ft;i++){
		//if( gft[i].name.length() != str.length()+1 ) continue;						//Why this assertion??
		
		for(j=0;j<str.length();j++){
			if( str.at(j)=='-' || str.at(j)=='_' || str.at(j)==',' ) continue;
			if( str.at(j) != gft[i].name.at(j) || j >= gft[i].name.length() ) break;
		}
		
		if( j == str.length() || j == gft[i].name.length() ) return i;					//found a exactly same one;
	}

	return -1;
}


/*
 * 1. This procedure initilize most of the data structure, excpet SAS+ forumulism;
 */
void Task::initialize_everything(){

	char command[1000];
	unsigned i,j,ft;	
	sprintf( command, "./translate/translate.py  %s %s ", 
						solverContext->domain_file, 
						solverContext->prob_file);
	system( command );	
	system("./preprocess/preprocess <output.sas");
	if( solverContext->debug )
		cout<<"./preprocess/preprocess done."<<endl;

	parse_fast_downward();
	//parse_fast_downward is suppose to run first, building v_mapping structure;
	parse();
	transGraphNum = graphGroup.size();
	
	for(unsigned g=0 ; g<mutexGroup.size() ; g++ ){
		vector<unsigned>& group = mutexGroup[g];
		for( i = group.size() -1 ; i > 0 ;i-- )
    	for( j = 0; j < i; j++){
			if( group[j] > group[j+1]){
				unsigned tmp = group[j];
				group[j] = group[j+1];
				group[j+1] = tmp;
			}
    	}
		
		for(i=0;i<group.size();i++)
		for(j=0;j<group.size();j++)
			if(i!=j){
				mutex_set.insert( make_pair(group[i],group[j]) );
				mutex_set.insert( make_pair(group[j],group[i]) );
			}
	}
	
	for(i=0;i<gnum_op;i++){
		sort( gop[i].P, gop[i].P + gop[i].num_P, less<int>());
		sort( gop[i].A, gop[i].A + gop[i].num_A, less<int>());
		sort( gop[i].D, gop[i].D + gop[i].num_D, less<int>());
	}
	
	struct timeb tp;
  	ftime( &tp );
  	srandom( tp.millitm );
	
	vector<unsigned>* vec_P = new vector<unsigned>[gnum_ft + transGraphNum];
	vector<unsigned>* vec_A = new vector<unsigned>[gnum_ft + transGraphNum];
	vector<unsigned>* vec_D = new vector<unsigned>[gnum_ft + transGraphNum];
	
	for(i=0;i<gnum_op;i++){

		for(j=0;j<gop[i].num_P;j++){
			ft = gop[i].P[j];
			//gft[ft].P[gft[ft].num_P++] = i;
			vec_P[ft].push_back(i);
		}

		for(j=0;j<gop[i].num_A;j++){
			ft = gop[i].A[j];
			vec_A[ft].push_back(i);
		}
		
		for(j=0;j<gop[i].num_D;j++){
			ft = gop[i].D[j];
			//gft[ft].D[gft[ft].num_D++] = i;
			vec_D[ft].push_back(i);
		}
	}
	
	for(i=0;i<gnum_ft + transGraphNum ;i++){
		gft[i].P = (unsigned*)malloc( sizeof(unsigned) * (vec_P[i].size() + 2 ) );	
		gft[i].num_P = vec_P[i].size();
		for(j=0;j<vec_P[i].size();j++) gft[i].P[j] = vec_P[i][j];
			
		gft[i].A = (unsigned*)malloc( sizeof(unsigned) * vec_A[i].size() );
		gft[i].num_A = vec_A[i].size();
		for(j=0;j<vec_A[i].size();j++) gft[i].A[j] = vec_A[i][j];
		
		gft[i].D = (unsigned*)malloc( sizeof(unsigned) * vec_D[i].size() );
		gft[i].num_D = vec_D[i].size();
		for(j=0;j<vec_D[i].size();j++) gft[i].D[j] = vec_D[i][j];
		
		gft[i].rand = random() % BIG_INT;
	}
	
	
	//Generate redundent facts (casued by NONE-x fact in DTGs);
	for(i=0;i<graphGroup.size();i++){
		Fact& fact = gft[ i + gnum_ft ];
		stringstream ss ;
		ss << i;
		fact.name = "NONE-" + ss.str(); 	//to-do: convert number into str;
	}
	
	delete[] vec_P;
	delete[] vec_A;
	delete[] vec_D;
}


void Task::analysis(){
	
	ordering_analysis();
	accompany_relations_analysis();			//Requires ordering info;
	goal_implications();
	goal_agenda_analysis(); 	

}

void Task::ordering_analysis(){
	cout<<"Analyzing Fact ordering... ";
	unsigned tf,i,j,k;
	bool mutex_set[MAX_FT_NUM];
	fact_ordering = new bool*[gnum_ft];
	bool**& ordering = fact_ordering;
	for(i=0;i<gnum_ft;i++){
		ordering[i] = new bool[gnum_ft];
		memset( ordering[i], false, sizeof(bool)*gnum_ft );
	}
	
	for( tf = 0 ;tf<gnum_ft ; tf++){
		if( gft[tf].graph == UNDEFINED ) continue;

		//build false set;
		memset( mutex_set, false, sizeof(bool) * gnum_ft);
		for(i=0;i<mutexGroup.size();i++){
			vector<unsigned>& group = mutexGroup[i];
			if ( bsearch_vector( group, tf ) != -1 ){
				for(j=0;j<group.size();j++)
					if( group[j] != tf && gft[group[j]].graph!=UNDEFINED)
						mutex_set[group[j]] = true;
			}
		}

		//test if $i<$tf holds;
		for( i =0;i<gnum_ft;i++){
			if( i == tf  || gft[i].graph == UNDEFINED) continue;
			
			//for all actions that add i;
			bool found = false;			
			for(j=0;j<gft[i].num_A;j++){
				unsigned oper = gft[i].A[j];

				for(k=0;k<gop[oper].num_D;k++)
					if ( tf == gop[oper].D[k]) break;
				if( k < gop[oper].num_D )
					continue;
					
				//print_op_name( oper ); cout<<" be checking"<<endl;									
				for(k=0;k<gop[oper].num_P;k++){
					unsigned pre = gop[oper].P[k];
					if( pre >= gnum_ft ) continue;
					
					if( mutex_set[pre] ){
						//print_ft_name( pre ); cout<<" is the qulified fact"<<endl;
						//ordering[i][tf] = true;
						break;
					}
				}

				if( k == gop[oper].num_P ){
					found = true; //At least possible-path found;
					break;
				}
			}
			
			if( !found )
				ordering[i][tf] = true;
		}
	}
	
	for(i=0;i<gnum_ft;i++)
	for(j=i;j<gnum_ft;j++)
		if( ordering[i][j] && ordering[j][i] )
			ordering[i][j] = ordering[j][i] = false;
	
	if( solverContext->display_info == 10){
		for(i=0;i<gnum_ft;i++)
		for(j=0;j<gnum_ft;j++)
			if( ordering[i][j] ){
				//cout<< i;
				print_ft_name( i );
				printf( "\t\t<_   \t");
				//cout<<j;
				print_ft_name( j );
				cout<<endl;
			}
		exit(0);
	}
	cout<<"done."<<endl;
}

/**
 * 	Part of the code in this function is from Metric-FF
 */
void Task::goal_agenda_analysis(){
	if( !solverContext->dgo  ) return;
	
	cout<<"Analyzing goal agenda... ";
	
	int i, j, k, n = goal.size(), start, entry;
	int *degree;
	int *hits;
	int *slot;

	degree = ( int * ) calloc( n, sizeof( int ) );
	hits = ( int * ) calloc( n, sizeof( int ) );
	slot = ( int * ) calloc( n, sizeof( int ) );
	for ( i = 0; i < n; i++ ) {
		degree[i] = 0;
		hits[i] = 0;
	}

	/* compute transitive closure on orderings matrix
	*/
	for ( j = 0; j < n; j++ ) {
		for ( i = 0; i < n; i++ ) {
			unsigned gi = goal[i];
			unsigned gj = goal[j];
			if ( fact_ordering[gi][gj] ) {
				for ( k = 0; k < n; k++ ) {
				  unsigned gk = goal[k];
				  if ( fact_ordering[gj][gk] ) 
					  fact_ordering[gi][gk] = true;
				}
			}
		}
	}

	/* count in - and outgoing edges, know those
	* goals that are not connected at all
	*/
	for ( i = 0; i < n; i++ ) {
		for ( j = 0; j < n; j++ ) {
		  if ( i != j && fact_ordering[goal[i]][goal[j]] ) {
			degree[i]--;
			degree[j]++;
			hits[i]++;
			hits[j]++;
		  }
		}
	}
	
	/* order goals with increasing degree, disconnected
   * at the very end.
   */
	for ( i = 0; i < n; i++ ) {
		if ( hits[i] == 0 ) {
			slot[i] = i;
			continue;
		}
	    for ( j = 0; j < i; j++ ) {
	      if ( degree[i] < degree[slot[j]] || hits[slot[j]] == 0 )
			break;
	    }
	    for ( k = i - 1; k >= j; k-- )
	      slot[k+1] = slot[k];
	    slot[j] = i;
	}
	
	start = 0;
	entry = 0;
	for ( i = 1; i < n; i++ ) {
	    if ( ( degree[slot[i]] != degree[slot[i-1]] ) ||
		 ( hits[slot[i]] == 0 && hits[slot[i-1]] != 0 ) ) {
		      //ggoal_agenda[entry].num_F = 0;
	    	  vector<unsigned> agenda;
		      for ( j = start; j < i; j++ ) {
				//ggoal_agenda[entry].F[ggoal_agenda[entry].num_F++] = 
				//  ggoal_state.F[slot[j]];
		    	  agenda.push_back( goal[slot[j]]);
		      }
		      goals_agenda.push_back( agenda );
		      start = i;
	    }
	}
	vector<unsigned> agenda;
	for ( i = start; i < n; i++ ) {
		agenda.push_back( goal[slot[i]]);
	}
	goals_agenda.push_back( agenda );
	
	
	vector<unsigned> vec;
	for(i=0;i<(int)goals_agenda.size();i++){
		for(j=0;j<(int)goals_agenda[i].size();j++){
			unsigned fact = goals_agenda[i][j];
			vec.push_back(fact);
		}
	}
	if( solverContext->debug == 0 ){
		for(i=0;i<(int)goal.size();i++){
			for(j=0;j<(int)vec.size();j++)
				if( goal[i] == vec[j] )
					break;
			if( j == (int)vec.size() )
				assert( false );		//Every goal is supposed to be exist in vec;
		}
	}
	vec = goal;
	
	if( solverContext->display_info == 13 ){
		cout<<endl<<"Goal Agendas:"<<endl;
		for(i=0;i<(int)goals_agenda.size();i++){
			cout<<"Agenda#"<<i<<": ";
			for(j=0;j<(int)goals_agenda[i].size();j++)
				print_ft_name( goals_agenda[i][j]),printf(";");
			cout<<endl;
		}
		exit(0);
	}
	
	free(degree);
	free(hits);
	free(slot);
	
	cout<<"  Done."<<endl;
}



void Task::goal_implications(){
	/*	debug
	for(unsigned i=0;i<mutexGroup.size();i++){
		
		for(unsigned j=0;j<mutexGroup[i].size();j++){
			print_ft_name( mutexGroup[i][j]);cout<<";";
		}
		cout<<endl;
	}
	
	set<pair<unsigned,unsigned> >::iterator iter = mutex_set.begin();
	for(;iter!=mutex_set.end();iter++){
		print_ft_name( iter->first );
		cout<<" and ";
		print_ft_name( iter->second );
		cout<<" are mutexes"<<endl;
	}*/
	
	cout<<"Deriving goals... ";
	
	unsigned gid,i,j,count,fact;
	bool* mutex = new bool[gnum_ft];
	memset( mutex, false, sizeof(bool) * gnum_ft );
	
	for(gid=0;gid<mutexGroup.size();gid++){
		
		if( mutexGroup[gid].size() <=3 ) 
			continue;	//Impossible to derive anything;
		for(i=0;i<mutexGroup[gid].size();i++)
			if( gft[mutexGroup[gid][i]].graph!= UNDEFINED )
				mutex[mutexGroup[gid][i]] = true;
	}
	
	for(i=0;i<goal.size();i++){
		
		for(gid=0;gid<mutexGroup.size();gid++){
			vector<unsigned>& group = mutexGroup[gid];
			int rtn = bsearch_vector( group, goal[i]);
			if( rtn == -1 ) continue;
			
			for(j=0;j<group.size();j++)
				mutex[group[j]] = false;
		}
	}
	
	for(gid=0;gid<mutexGroup.size();gid++){

		vector<unsigned>& group = mutexGroup[gid];
		
		for(i=0;i<group.size();i++){
			fact = group[i];
			if( !mutex[fact]) continue;
			
			vector<unsigned> test_set(goal);
			test_set.push_back( fact );
			if( detect_deadend(test_set) ){
				cout<<"Setting false";print_ft_name( fact );
				mutex[fact] = false;
				//If a set is found to be impossible, not necessary to detect another one
				//break;	
			}
		}
		//case of count = 1 will be taken care of later;
	}

/* debug
	cout<<"Facts that is still enabled"<<endl;
	for(gid=0;gid<transGraphNum;gid++){
		if( transGraph[gid].hasNegateAll ) 
			continue;
		int count = 0;
		for(i=0;i<transGraph[gid].size;i++){
			fact = transGraph[gid].vertex[i][0];
			if( mutex[fact] ) 
				print_ft_name(fact);
		}
		cout<<endl;
	}		//*/

	vector<unsigned> imp_g;
	for(gid=0;gid<mutexGroup.size();gid++){
		vector<unsigned>& group = mutexGroup[gid];	
		
		count = 0;
		for(j=0;j<group.size();j++)
			if( mutex[group[j]] )  count++;

		if( count !=1 ) continue;
		
		for(j=0;j<group.size();j++){
			if( !mutex[group[j]] ) continue;
			
			//To avoid adding duplicated ones;
			//This implication routine still seems buggy, need to be looked into; (Storage)
			vector<unsigned>::iterator where = find(imp_g.begin(), imp_g.end(), group[j]);
			if( where == imp_g.end() )
				imp_g.push_back( group[j] );
		}
	}
	
	if( solverContext->display_info == 12 ){
		cout<<imp_g.size()<<" facts implied to be goals."<<endl;
		cout<<"Implied goals:"<<endl;
		for(i=0;i<imp_g.size();i++)
			printf("         "),print_ft_name(imp_g[i]),printf("\n");
		cout<<endl;

		exit(0);
	}

	for(i=0;i<imp_g.size();i++){
		for(j=0;j<aco_pr_vec.size();j++)
			if( imp_g[i] ==  aco_pr_vec[j].first ) break;
		if( j == aco_pr_vec.size() )
			goal.push_back( imp_g[i] );
	}
	
	delete[] mutex;
	cout<<" done."<<endl;
}



