/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/


#include<vector>
#include<iostream>
#include<fstream>
#include<sstream>
#include<cassert>
#include<stdio.h>
#include<stdlib.h>

#include "dp.h"
#include "state.h"
#include "plan.h"
#include "transition_graph.h"
#include "task.h"
#include "solver_context.h"
#include "search_entities.h"
#include "./fd/fd_globals.h"
#include "utils.h"

using namespace std;


//Preconditions that must be true before fact can be added;
vector<unsigned> Task::get_enforced_preconditions(unsigned fact){
	unsigned i,j,k,l,op,firstOp;
	vector<unsigned> result;
	
	if( gft[fact].num_A == 1 ){
		op = gft[fact].A[0];
		for(i=0;i<gop[op].num_P;i++)
			result.push_back( gop[op].P[i] );
		return result;
	}
	
	for(i=0;i<gft[fact].num_A;i++){
		firstOp = gft[fact].A[0];
		
		for(j=0;j<gop[firstOp].num_P;j++){
			unsigned checkingP = gop[firstOp].P[j];
			
			for( l=1; l<gft[fact].num_A;l++){
				op = gft[fact].A[l];
				
				for(k=0;k<gop[op].num_P;k++)
					if( checkingP == gop[op].P[k]) break;
				
				if( k == gop[op].num_P )
					break;
			}
			
			if( l == gft[fact].num_A )
				result.push_back( checkingP );
		}
	}

	/*
	cout<<"Forced Precondition of ";
	print_ft_name(fact);
	cout<<":"<<endl;
	for(i=0;i<result.size();i++)
		print_ft_name( result[i] ),printf(";");
	cout<<endl; */
	
	return result;
}


bool Task::detect_deadend( vector<unsigned> test_set){
	unsigned i,j,k,l,m,op;
						
	for( i=0; i<test_set.size(); i++){
		
		vector<unsigned> pre_vec = get_enforced_preconditions( test_set[i] );
		for(j=0;j<pre_vec.size();j++){

			for(k=0;k<test_set.size();k++){
				if( k == i) continue;
				//Condition#1;
				set<pair<unsigned,unsigned> >::iterator iter = 
									mutex_set.find( make_pair( pre_vec[j],test_set[k] ));
				if( iter == mutex_set.end() )
					continue;
				
				//Condition#2. test_set[k],test_set[i] cannot be added at the same time;
				//			   with pre_vec[j] as a precondition;
				for(l=0;l<gft[pre_vec[j]].num_P;l++){
					unsigned flag1 = false,flag2 = false;
					op = gft[pre_vec[j]].P[l];
					
					for(m=0;m<gop[op].num_A;m++)
						if ( gop[op].A[m] == test_set[k])
							flag1 = true;
						else if ( gop[op].A[m] == test_set[i] )
							flag2 = true;
					if( flag1 && flag2 ) break;
				}
				
				if( l == gft[pre_vec[j]].num_P ) break;
			}
			
			if( k < test_set.size() ) 
				break;
		}
		
		if( j == pre_vec.size() ) return false;
	}
	return true;
}


void Task::accompany_relations_analysis(){
	assert( fact_ordering != NULL );
	
    unsigned i,j,k,l;
    unsigned** pr = (unsigned **)malloc( sizeof(unsigned *) * gnum_ft );
    for(i=0;i< gnum_ft;i++){
        pr[i] = (unsigned*)malloc( sizeof(unsigned) *  gnum_ft);
        memset( pr[i], 0, sizeof(unsigned) * gnum_ft );
    }
    
    for(i=0;i< gnum_ft;i++){

        vector<unsigned> candidate_list;
        
        if( gft[i].num_A > 0 ){
            int firstOp = gft[i].A[0];
            
            for(j=0;j<gop[firstOp].num_A;j++){
                unsigned candidate = gop[firstOp].A[j];
                if( i ==  candidate ) continue;
                
                //Check all the actions except the first one;
                for(k=1;k<gft[i].num_A;k++){
                    unsigned op = gft[i].A[k];
                    for(l=0;l<gop[op].num_A;l++)
                        if ( candidate == gop[op].A[l] ) break;
                    
                    if( l == gop[op].num_A) break; //not found then break;
                }
                
                //Meaning the candidate if valid;
                if( k == gft[i].num_A )
                    candidate_list.push_back( candidate );
            }
        }
        //There must be another condition says that nothing can change them(i.e. i and candidate)!!!!!!!!!!!!!!1
        
        if( gft[i].num_D > 0 ){
            unsigned firstOp = gft[i].D[0];
            for(j=0;j<gop[firstOp].num_D;j++){
                unsigned candidate = gop[firstOp].D[j];
                if( i == candidate ) continue;
                
                //Must obey the condition of Add-effect above;
                for(k=0;k<candidate_list.size();k++)
                    if( candidate_list[k] == candidate ) break;
                if( k ==candidate_list.size() ) continue;
                
                for( k=1;k<gft[i].num_D;k++){
                    int op = gft[i].D[k];
                    for(l=0;l<gop[op].num_D;l++)
                        if ( candidate == gop[op].D[l] ) break;
                    
                    if( l == gop[op].num_D ) break;
                }
                
                if( k == gft[i].num_D )
                    pr[i][candidate]  =1 ;
            }
        }
    }


    for(i=0;i<gnum_ft;i++){
        vector<unsigned> vec;
        for(j=0;j<gnum_ft;j++)
            if( pr[j][i] )
                vec.push_back(j);
        
        if( vec.size() > 0){
            aco_pr_vec.push_back( make_pair( i, vec ) );
        }
    }
    
    
   	//Sort all slave facts;
   	for(i=0;i<aco_pr_vec.size() ;i++){
   		vector<unsigned>& slaves = aco_pr_vec[i].second;
   		for( k = slaves.size()-1 ; k > 0 ;k-- )
   		for(j = 0; j < k; j++){
   			assert( slaves[j+1]<gnum_ft && slaves[j] < gnum_ft );
   			if( fact_ordering[slaves[j+1]][slaves[j]]  ){
   				unsigned tmp = slaves[j];
   				slaves[j] = slaves[j+1];
   				slaves[j+1] = tmp;	
   			}
   		}
   	}
   	
    //Debug Output:
    if( solverContext->display_info == 11 ){
	    printf("Accompany Fact Relations:\n");
	    for(i=0;i<aco_pr_vec.size();i++){
	        print_ft_name( aco_pr_vec[i].first );
	        vector<unsigned>& slaves = aco_pr_vec[i].second;
	        printf(":\n");
	        for(j=0;j<aco_pr_vec[i].second.size();j++)
	            printf("          "),print_ft_name(slaves[j]),printf("\n");
	    } 
	    exit(0);
    }
    
    for(i=0;i< gnum_ft;i++) free(pr[i]);
    free(pr);
}


//Only for accompanying goals solving strategy use;
bool Task::test_reachability( 	State& state, 
								vector<unsigned> fixed_facts, 
								unsigned goal )
{
	
	vector<unsigned> disabled_facts;
	unsigned i,j,fact,oper,asgn;
	
	for(i=0;i<gnum_ft;i++)
	for(j=0;j<fixed_facts.size();j++){
		MutexSetType::iterator iter = mutex_set.find( make_pair(i,fixed_facts[j]));
		if( iter!= mutex_set.end()){
			disabled_facts.push_back( i);
			break;
		}
	}
/*	
	cout<<"Disabled fact:"<<endl;
	for(i=0;i<disabled_facts.size();i++)
		print_ft_name( disabled_facts[i]),cout<<";";
	cout<<endl; //*/
	
	vector<unsigned> dis_ops;
	vector<unsigned> app_ops;
	for(i=0;i<disabled_facts.size();i++){
		fact = disabled_facts[i];
		for(j=0;j<gft[fact].num_P;j++)
			dis_ops.push_back( gft[fact].P[j]);
	}
	sort( dis_ops.begin(), dis_ops.end(), less<unsigned>());
	for(i=0;i<gnum_op;i++)
		if ( bsearch_vector(dis_ops, i) == -1 )
			app_ops.push_back(i);
	sort( app_ops.begin(), app_ops.end(), less<unsigned>() );
	
	set<unsigned> r_state;
	for(i=0;i<transGraphNum;i++){
		asgn = state[i];
		if( !transGraph[i].vertex[asgn].isNone() )
			r_state.insert( transGraph[i].vertex[asgn][0] );
	}
	
	set<unsigned>::iterator rs_iter;
	while(true){
		bool flag = false;
		vector<unsigned>::iterator iter = app_ops.begin();
		
		for( ;iter!=app_ops.end();){
			oper = *iter;
			
			for(i=0;i<gop[oper].num_P;i++){
				rs_iter = r_state.find( gop[oper].P[i] );
				if( rs_iter == r_state.end() ) break;
			}
			
			if( i < gop[oper].num_P){
				iter++;
				continue;
			}
			
			flag = true;
			//cout<<"Applying action"; print_op_name( oper );cout<<endl;
			for( i=0 ; i<gop[oper].num_A ; i++ )
				r_state.insert( gop[oper].A[i] );
			app_ops.erase( iter );
		}
		
		if( !flag ) break;
	}
	
	
	if( r_state.find( goal ) ==  r_state.end() )  return false;
	
	return true;
}










/**
 * 
 * 				Outputs 
 * 
 */


void Task::print_goals(){
	unsigned i;
	cout<<"Goals: total "<<goal.size() <<endl;
	for(i=0;i<this->goal.size();i++){
		gft[goal[i]].print_name();
		cout<<";";
	}
	cout<<endl;
}

void Task::print_types(){
	unsigned i;
	cout<<"Types:"<<endl;
	for(i=0;i<types.size();i++)
		cout<<"  "<<types[i]<<endl;
}	

void Task::print_all_gops(){
	unsigned i;
	cout<<"Objects"<<endl;
	for(i=0;i<objects.size();i++)
		cout<<"  "<<objects[i].name<<endl;
}

void Task::print_all_gfts(){
	unsigned i;
	cout<<"Facts"<<endl;
	for(i=0;i<this->gnum_ft;i++)
		cout<<"  "<<this->gft[i].name<<endl; 
}


void Task::print_ft_name(unsigned g, unsigned v){
	assert( transGraph[g].dimension == 1 );
	gft[transGraph[g].vertex[v][0]].print_name();
}


void Task::print_ft_name(unsigned ft){ 
	gft[ft].print_name();
}

void Task::print(){
	
}


