/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cassert>
#include <stdio.h>
#include <stdlib.h>

#include "dp.h"
#include "state.h"
#include "plan.h"
#include "task.h"
#include "transition_graph.h"
#include "solver_context.h"
#include "search_entities.h"
#include "./fd/fd_globals.h"
#include "./fd/fd_operator.h"
#include "utils.h"

using namespace std;


void Task::parse(){
	assert( solverContext != NULL );
	
	ifstream ifs("problem.definition");
	assert(ifs!=NULL);
	
	unsigned i,j,count,num;
	int re;
	string str;

	//Types
	ifs>>count;
	getline(ifs,str);	
	for(i=0;i<count;i++){
		getline( ifs, str );
		this->types.push_back( str );
	}
	
	//Objects
	ifs>>count;
	getline(ifs,str);
	for(i=0;i<count;i++){
		getline(ifs, str );
		Object obj(str);
		this->objects.push_back( obj );
	}
	
	//predicates;
	ifs>>count;
	getline(ifs,str);	
	for(i=0;i<count;i++){
		getline( ifs, str );
		this->predicates.push_back( str );
	}
		
	//Facts
	ifs>>count;
	getline(ifs,str);
	this->gft = new Fact[ count + MAX_DTG_NUM ];
	this->gnum_ft = count;
	for(i=0;i<count; i++){
		getline( ifs, str );
		int tmp = str.find(" ");
		this->gft[i].name = str.substr(tmp);
		while( (re = this->gft[i].name.find(" "))!=-1)
			this->gft[i].name = gft[i].name.substr(0,re) + 
					gft[i].name.substr( re+1, gft[i].name.length() );
	}
	
	//Initial State;
	ifs >> count;
	for(i=0;i<count;i++){
		ifs>>num;
		this->init.push_back(num);
	}

	//Goal;
	ifs >> count;
	for(i=0;i<count;i++){
		ifs >> num;
		this->goal.push_back(num);
		this->orig_goals.push_back(num);
	}
	sort( orig_goals.begin(), orig_goals.end(), less<unsigned>());
	
	//DTG
	ifs >> count;
	assert( count < MAX_DTG_NUM );
	for(i=0;i<this->var_mapping_count;i++){
		vector<unsigned> empty_vec;
		this->graphGroup.push_back(empty_vec);
	} 
	for(i=0;i<count;i++){
		ifs >> num;
		vector<unsigned> tmp_vec;
		for( j=0; j<num; j++ ){
			ifs >> re;
			
			if( re == -1 )
				tmp_vec.push_back( UNDEFINED );
			else
				tmp_vec.push_back( re );
		}
		//If the var$ is removed by downward's preprocessing, then we also ignore it;
		if( var_mapping[i] != UNDEFINED )
			graphGroup[var_mapping[i]] = tmp_vec;
	}
	
	//Handle Exception: Some facts doesn't exists in any graph Group, add each of them individually;
	bool* flag = new bool[gnum_ft];
	memset(flag, false, sizeof(bool) * gnum_ft );
	for(i=0;i<graphGroup.size();i++){
		for(j=0;j<graphGroup[i].size();j++){
			if(graphGroup[i][j]!=UNDEFINED)
				flag[graphGroup[i][j]] = true;
		}
	}
	for(i=0;i<gnum_ft;i++)
		if( !flag[i] ){			
			vector<unsigned> tmp_vec;
			tmp_vec.push_back( i );
			tmp_vec.push_back( UNDEFINED );
			graphGroup.push_back(tmp_vec);
		}

	delete[] flag;
	//End Exception;
	
	
	//Mutex Group
	ifs >> count;
	for(i=0;i<count;i++){
		ifs >> num;
		vector<unsigned> tmp_vec;
		for( j=0; j<num; j++ ){
			ifs >> re;
			tmp_vec.push_back(re);
		} 
		mutexGroup.push_back( tmp_vec );
	}

	
	//Operators;
	ifs >> count;
	this->gop = new Operator[count+FAST_DOWNWARD::g_axioms.size() + 10];
	this->gnum_op = count;
	for(i=0;i<count;i++)
		this->gop[i].parse(ifs);

	ifs.close();

	for(i=0;i<gnum_ft;i++){
		gft[i].predicate = UNDEFINED;
		for(j=0;j<predicates.size();j++){
			int rtn = strncmp( 	predicates[j].c_str() , 
								gft[i].name.c_str() , 
								strlen( predicates[j].c_str())  );
			if( rtn == 0 ){
				gft[i].predicate = j;
				break;
			}
		}
	}
	
	
	if( solverContext->display_info == 7 ){
		cout<<"Mutex Groups:"<<endl;
		for(i=0;i<mutexGroup.size();i++){
			cout<<"#"<<i<<":";
			for(j=0;j<mutexGroup[i].size();j++)
				print_ft_name( mutexGroup[i][j] );
			cout<<endl;
		}
		exit(0);
	}
	
	if( solverContext->display_info == 8 ){
		
		cout<< graphGroup.size()<<"graph groups in total."<<endl;
		for(i=0;i<graphGroup.size();i++){
			cout<<"#"<<i<<graphGroup[i].size()<<":";
			for(j=0;j<graphGroup[i].size();j++){
				if(graphGroup[i][j]!=UNDEFINED)
					print_ft_name( graphGroup[i][j] ),printf(" ");
				else
					cout<<" n/a ";
			}
			cout<<endl;
		}
		exit(0);
	}
}

/*
 * 	This procedure may not be stable.
 */
typedef pair<unsigned,pair<unsigned,unsigned> > PrePost;
void Task::parse_axioms(){
	unsigned i,j,k,gid;
		
	vector<FAST_DOWNWARD::Operator>& gaxioms = FAST_DOWNWARD::g_axioms;
	gpos_xm = gnum_op;

	//1.Parse axiom information from downward;
	for(i=0;i<gaxioms.size();i++){
		FAST_DOWNWARD::Operator& oper = gaxioms[i];
		vector<PrePost> vec_T;
		vector<unsigned> vec_A,vec_P,vec_D;
		
		for(j=0;j<oper.get_pre_post().size();j++){
			
			FAST_DOWNWARD::PrePost prepost = oper.get_pre_post()[j];
			for(k=0;k<prepost.cond.size();k++){
				//VAR:prepost.cond[k].var & PREVAIL prepost.cond[k].prev;
				PrePost unit = make_pair( prepost.cond[k].var, 
						make_pair(prepost.cond[k].prev,prepost.cond[k].prev));
				vec_T.push_back(unit);
			}
			//VAR prepost.var & PREVAIL prepost.pre --> prepost.post
			PrePost unit = make_pair( prepost.var, 
								make_pair( prepost.pre, prepost.post));
			vec_T.push_back( unit );
			assert( prepost.var >=0 && (unsigned)prepost.var < transGraphNum );
			transGraph[ prepost.var].isAxiom = true;
		}
		
		Operator& axiom = gop[gpos_xm];

		axiom.Tg = (unsigned*) malloc( sizeof(unsigned) * vec_T.size() );
		axiom.Tp = (unsigned*) malloc( sizeof(unsigned) * vec_T.size() );
		axiom.Te = (unsigned*) malloc( sizeof(unsigned) * vec_T.size() );
		axiom.num_T = vec_T.size();
		
		for(j=0;j<vec_T.size();j++){

			PrePost& unit = vec_T[j];
			axiom.Tg[j] = unit.first;
			axiom.Tp[j] = unit.second.first;
			axiom.Te[j] = unit.second.second;
			
			if( axiom.Tp[j] == axiom.Te[j] )
				vec_P.push_back( transGraph[axiom.Tg[j]].vertex[axiom.Tp[j]][0]);
			else{
				vec_D.push_back(transGraph[axiom.Tg[j]].vertex[axiom.Tp[j]][0]);
				vec_A.push_back(transGraph[axiom.Tg[j]].vertex[axiom.Te[j]][0]);
				axiom.Tp[j] = NO_PRE_NEEDED;						//Hack ..., Let's see if it's correct...				
			}
		}

		axiom.P  = (unsigned*) malloc( sizeof(unsigned) * vec_P.size() );
		axiom.D  = (unsigned*) malloc( sizeof(unsigned) * vec_D.size() );
		axiom.A  = (unsigned*) malloc( sizeof(unsigned) * vec_A.size() );
		axiom.num_P = vec_P.size();
		axiom.num_A = vec_A.size();
		axiom.num_D = vec_D.size();
		
		for(j=0;j<vec_P.size();j++) axiom.P[j] = vec_P[j];
		for(j=0;j<vec_D.size();j++) axiom.D[j] = vec_D[j];
		for(j=0;j<vec_A.size();j++) axiom.A[j] = vec_A[j];
		gpos_xm++;
	}
	
	//to-do:Remove NP in each operator;
	for( i = 0 ; i < gnum_op ; i ++ ){
	}
	
	//2.Re-calculate the axiom-related graphs' edges; 
	//It should be necessary, otherwise, the transition in the axiom-related-graphs would
	//never get satisfied;
	for( gid = 0 ; gid < transGraphNum; gid++ ){
		if ( !transGraph[gid].isAxiom ) continue;
		TransitionGraph& graph = transGraph[gid];
		
		for(i=0;i<graph.size;i++)
		for(j=0;j<graph.size;j++)
			if( i!=j && !graph.vertex[i].isNone() && graph.vertex[j].isNone() ){
				
				vector<unsigned> vec_op;
				unsigned axiom = UNDEFINED;
				for(axiom = gnum_op; axiom<gpos_xm ; axiom++ )
					for(k=0;k<gop[axiom].num_T;k++)
						if( gop[axiom].Tg[k] == gid && 
							gop[axiom].Tp[k] == i && 
							gop[axiom].Te[k] == j){
							vec_op.push_back( axiom );
							break;
						}
				
				graph.edge[i][j].action = 
					(unsigned*) malloc( sizeof(unsigned) * vec_op.size() );
				graph.edge[i][j].num_action = vec_op.size();
				for(k=0;k<vec_op.size();k++)
					graph.edge[i][j].action[k] = vec_op[k];
			}
	}
	
	
	cout<<"Totally "<<gpos_xm - gnum_op <<" axioms parsed."<<endl;

	if( solverContext->display_info == 9 ){
		
		unsigned axiom;
		for( axiom = gnum_op ; axiom < gpos_xm; axiom++ ){
			cout<<"P:";
			for( i = 0; i< gop[axiom].num_P;i++)
				print_ft_name( gop[axiom].P[i] );
			cout<<"EF:";
			for( i=0;i<gop[axiom].num_A; i++)
				print_ft_name( gop[axiom].A[i] );
			cout<<endl;
		}
		
		exit(0);
	}
}

void Task::parse_fast_downward(){
	
	const int PARSE_FD_DEBUG = 0; 
	
	string input_file = "output";
	filebuf fb;
	fb.open ( input_file.c_str() , ios::in );
	istream is(&fb);
	
	int poly_time_method;
	is >> poly_time_method;
	FAST_DOWNWARD::read_everything( is );
	
	vector<string>& var_name_list = FAST_DOWNWARD::g_variable_name;
	memset( var_mapping, UNDEFINED, sizeof(unsigned)* MAX_DTG_NUM );
	
	var_mapping_count = var_name_list.size();
	if( PARSE_FD_DEBUG ) cout<<"var_mapping_count"<<var_mapping_count<<endl;
	
	for(unsigned i=0;i<var_name_list.size();i++){
		//cout<<var_name_list[i]<<endl;debug
		
		string number = var_name_list[i].substr( 3, var_name_list[i].size()-3 );
		int num = atoi(number.c_str());
		if( PARSE_FD_DEBUG ) 
			cout<<"PARSE_FD_DEBUG:"<<var_name_list[i]<<":"<<number<<"AS"<<num<<endl;
		
		var_mapping[ num ] = i;
	}
	
	if( PARSE_FD_DEBUG ){
		for(unsigned i=0;i<var_mapping_count;i++){
			if( var_mapping[i] == UNDEFINED  )
				cout<<"UNDEFINED"<<endl;
			else
				cout<<"#"<<i<<":"<<var_mapping[i]<<endl;
		}
	}
}



