/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>


#include "dp.h"
#include "state.h"
#include "plan.h"
#include "transition_graph.h"
#include "task.h"
#include "solver_context.h"
#include "search_entities.h"
#include "./fd/fd_globals.h"
#include "utils.h"

using namespace std;

/*
 * 
 * Assumptions:	
 * 	#1. Since in the preprocessing stage, 
 * 		All the graphs involved are those size-one-vertex graphs;
 * 
 * *When calling this, the basic structure of all the graphs are ready;
 */
void Task::action_rep_transform(){

	unsigned i,g,j,l,k,m,action,nT;
	
	for(i=0;i<gnum_op;i++){ 
		int estimateNumber = (gop[i].num_P + gop[i].num_A + gop[i].num_D ) * 10;
		gop[i].Tg = (unsigned *)malloc( sizeof(unsigned) * estimateNumber );
		gop[i].Tp = (unsigned *)malloc( sizeof(unsigned) * estimateNumber );
		gop[i].Te = (unsigned *)malloc( sizeof(unsigned) * estimateNumber );
		gop[i].num_T = 0;
	}
	
	//Deal with those CHANGES in transitions;
	for(g=0;g<transGraphNum;g++){
		TransitionGraph& graph = transGraph[g];
		
		assert( graph.edge!= NULL );
		
		for(i=0;i<graph.size;i++)
		for(j=0;j<graph.size;j++){
			
			TransitionEdge& edge = graph.edge[i][j];
			
			if( edge.num_action == 0 ) continue;
			
			for(l=0;l<edge.num_action;l++){
				action = edge.action[l];
				nT = gop[action].num_T;
				
				//Make sure $i is one of action's precondition
				bool flag = false;
				if( graph.vertex[i].isNone() )
					//The change here is for converting action representation into downward's SAS output
					flag = false ;// it used to be "= true "to-do: may need to check if it's axiom;
				else{
					for(m=0;m<gop[action].num_P;m++)
						if( gop[action].P[m] == graph.vertex[i][0] ){
							flag = true;
							break;
						}
				}
				
				gop[action].Tg[nT] = g;				
				gop[action].Tp[nT] = flag ? i : NO_PRE_NEEDED;
				gop[action].Te[nT] = j;
				gop[action].num_T++;
			}
		}
	}

	//Handle:
	//	1.Those pre-conditions that are not related with any effects;
	//	2.Those effects(add&delete) that haven't been paired with any precondition;
	unsigned pre = UNDEFINED;
	for(i=0;i<gnum_op;i++){
		if( gop[i].num_T == 0 ) continue; //Dummy actions
		if( gop[i].isDisabled ) continue;
		
		bool flag = false;
		for(j=0;j<gop[i].num_P;j++){
			pre = gop[i].P[j];
			if( gft[pre].vertex == UNDEFINED && gft[pre].graph == UNDEFINED){
				gop[i].isDisabled = true;
				flag = true;
				break;
			}
		}
		if( flag ) continue;
		
		for(j=0;j<gop[i].num_P;j++){
			pre = gop[i].P[j];
			assert( pre >= 0 && pre < gnum_ft );
			for(k=0;k<gop[i].num_T;k++){
				if ( gft[pre].graph == gop[i].Tg[k] )
					goto doNothing;
			}
			
			assert( gft[pre].vertex != UNDEFINED );
			
			nT = gop[i].num_T;
			gop[i].Tg[nT] = gft[pre].graph;
			gop[i].Tp[nT] = gft[pre].vertex;
			gop[i].Te[nT] = gft[pre].vertex;
			gop[i].num_T++;

			assert(  gop[i].num_T < (gop[i].num_P + gop[i].num_A + gop[i].num_D)* 10 );
	doNothing:
			continue;
		}
		
		vector<unsigned> to_remove_NP;
		/// This is mostly for handling "NewAxiom" atom, which is always a negation;
		for( j = 0; j< gop[i].num_NP ; j++){
			unsigned pre = gop[i].NP[j]; 
			assert(  pre >=0 && pre < gnum_ft );
			
			int cpn= strncmp("new-axiom@", gft[pre].name.c_str() ,strlen("new-axiom@"));
			//It is a hack.... to make the axioms work;
			if( cpn == 0 ){
				unsigned gid = gft[pre].graph;
				unsigned vid = gft[pre].vertex;
				
				assert( vid == 0 );
				assert( transGraph[gid].size == 2 );
				
				unsigned nT = gop[i].num_T;
				gop[i].Tg[nT] = gid;
				gop[i].Tp[nT] = 1;
				gop[i].Te[nT] = 1;
				gop[i].num_T++;
				
				gop[i].P[gop[i].num_P++] = transGraph[gid].vertex[1][0];
				to_remove_NP.push_back(j);
			}
			
			/*else{
				//This version doesn't support negated precondition;
				cout<<"Caution:Not AxiomAtom!!!Exception."<<endl;
				print_ft_name( pre );
				exit(0);
			} */
		}
	}
	
	for(i=0;i<gnum_op;i++){   
		gop[i].Tg = (unsigned *)realloc( gop[i].Tg, sizeof(unsigned) * gop[i].num_T );
		gop[i].Tp = (unsigned *)realloc( gop[i].Tp, sizeof(unsigned) * gop[i].num_T );
		gop[i].Te = (unsigned *)realloc( gop[i].Te, sizeof(unsigned) * gop[i].num_T );
	}
	
	printf("Transforming action representation to SAS+... Done.\n");fflush(stdout);
}



void Task::build_sas_formulation(){
	unsigned i,g;
	
	//Step1.build transition graphs
	transGraph = new TransitionGraph[graphGroup.size()];
	for(i=0;i<graphGroup.size();i++){
		transGraph[i].id = i;
		transGraph[i].build_vertexes( graphGroup[i] );
		transGraph[i].ptr_task = this;
	}
	
	//Step2.Build mapping between gft_conn and DTGs;
	for(i=0;i<gnum_ft + transGraphNum;i++){
		gft[i].graph = UNDEFINED;
		gft[i].vertex = UNDEFINED;
	}
	
	for(g=0;g<transGraphNum;g++){
		TransitionGraph& graph = transGraph[g];
		for(i=0;i<graph.size;i++){
				
			int factId = graph.vertex[i][0];
			assert( gft[factId].graph == g || gft[factId].graph == UNDEFINED );
			gft[factId].graph = g;
			
			assert( graph.dimension == 1 );
			gft[factId].vertex = i;
		}
	}
	
	disable_unreachable_actions();
	
	//Step4. build edges in graphs
	for(i=0;i<transGraphNum;i++)
		transGraph[i].build_edges();
	
	//Step5. Translate actions' {P,A,D} info into {Tg,Tf,Te,Tp};
	action_rep_transform();
		
	//Step6. Build dependency relationships
	build_dependencies();
	
	//Step7.
	//Assumption: in such graphs, any transition edge has only one action;
	cout<<"calcing inner path/distances for all graphs..."<<flush;
	for(g=0;g<transGraphNum;g++){
		TransitionGraph& graph = transGraph[g];
		graph.path = NULL;
		for(i=0;i<transGraphNum;i++)
			if( causalMatrix[g][i] ) break;
		if( i<transGraphNum ) continue;
		
		graph.build_inner_path();
	}
	cout<<"  Done"<<endl;
	
	//Step 8.
	for(g=0;g<transGraphNum;g++){
		transGraph[g].calc_distances();
		transGraph[g].compute_identifier();
	}
	cout<<"Calc distancs and type identifiers Done"<<endl;
	
	//Step.9  When all done. Read in axioms;
	parse_axioms();	
	
	if( this->solverContext->display_info  == 100){
		for(i=0;i<transGraphNum;i++) 
			transGraph[i].print();
		exit(0);
	}
	
	cout<<"SAS+ formulism done."<<endl;
}


void Task::disable_unreachable_actions(){
	unsigned i,j,k;
	
	//Step3. Disable unreachable actions;
	for(i =0;i<gnum_ft;i++){
		//disable those actions related with un-reachable facts;
		if( gft[i].graph == UNDEFINED && gft[i].vertex == UNDEFINED){
			/*
			for( j = 0;j<gft[i].num_D ;j ++){
				unsigned oper = gft[i].D[j];
				gop[oper].isDisabled = true;				
			}
			for( j = 0;j<gft[i].num_P ;j ++){
				unsigned oper = gft[i].P[j];
				gop[oper].isDisabled = true;				
			}
			for( j = 0;j<gft[i].num_A ;j ++){
				unsigned oper = gft[i].A[j];
				gop[oper].isDisabled = true;				
			} //*/
			
			for( j = 0;j<gft[i].num_D ;j ++){
				unsigned oper = gft[i].D[j];
				for(k=0;k<gop[oper].num_D;k++)
					if( gft[gop[oper].D[k]].graph != UNDEFINED )
						goto Reachable;
				for(k=0;k<gop[oper].num_A;k++)
					if( gft[gop[oper].A[k]].graph != UNDEFINED )
						goto Reachable;
				gop[oper].isDisabled = true;
				//gop[oper].print_name();	cout<<"Found unreachable action"<<endl;
			}
			
			for( j = 0;j<gft[i].num_A ;j ++){
				unsigned oper = gft[i].A[j];
				for(k=0;k<gop[oper].num_D;k++)
					if( gft[gop[oper].D[k]].graph != UNDEFINED )
						goto Reachable;
				for(k=0;k<gop[oper].num_A;k++)
					if( gft[gop[oper].A[k]].graph != UNDEFINED )
						goto Reachable;
				gop[oper].isDisabled = true;
				//gop[oper].print_name();	cout<<"Found unreachable action"<<endl;
			}
			
		Reachable:
			continue;
		}
		
		assert( gft[i].vertex < transGraph[gft[i].graph].size );		
	}
	
	//Step3.b Cancel the unexpected `isDisabled' flags;(Rovers.)
	//			Might be a little expensive any way;
	
	for(i=0;i<gnum_op;i++){
		if( !gop[i].isDisabled ){
			for(j=0;j<gop[i].num_P;j++){
				if(gft[gop[i].P[j]].graph == UNDEFINED &&  gft[gop[i].P[j]].vertex == UNDEFINED){
					unsigned m;
					gop[i].num_P--;
					for(m=j;m<gop[i].num_P;m++) gop[i].P[m] = gop[i].P[m+1];
					j--;
				}
			}
			continue;
		}
		
		for(j=0;j<gop[i].num_P;j++){
			unsigned fact = gop[i].P[j];
			unsigned k,l,m;
			
			for(k=0;k<gop[i].num_D;k++)
				if( gop[i].D[k] == fact ) break;
			if( k == gop[i].num_D ) continue;
			
			for(l=0;l<gop[i].num_A;l++)
				if( gop[i].A[l] == fact ) break;
			if( l == gop[i].num_A ) continue;
			
			//$fact can be removed from op;
			gop[i].isDisabled = false;
			gop[i].num_D--;
			gop[i].num_A--;
			for(m=k;m<gop[i].num_D;m++) gop[i].D[m] = gop[i].D[m+1];
			for(m=l;m<gop[i].num_A;m++) gop[i].A[m] = gop[i].A[m+1];
			gop[i].num_P--;
			for(m=j;m<gop[i].num_P;m++) gop[i].P[m] = gop[i].P[m+1];
			j--;
		}
	}

	for(unsigned op=0;op<gnum_op;op++){
		for(i=0;i<gop[op].num_P;i++)
		for(j=0;j<gop[op].num_P;j++){
			if( i == j || gop[op].P[i] >= gnum_ft || gop[op].P[j] >= gnum_ft ) continue;
			
			set< pair<unsigned,unsigned> >::iterator iter = mutex_set.find(
						make_pair( gop[op].P[i] , gop[op].P[j] ) );
			if( iter!= mutex_set.end() ){
				gop[op].isDisabled = true;
				goto TEST_NEXT_OP;
			}
		}
		
		for(i=0;i<gop[op].num_A;i++)
		for(j=0;j<gop[op].num_A;j++){
			if( i == j || i > gnum_ft || j > gnum_ft ) continue;
			
			set< pair<unsigned,unsigned> >::iterator iter = mutex_set.find(
						make_pair( gop[op].A[i] , gop[op].A[j] ) );
			if( iter!= mutex_set.end() ){
				gop[op].isDisabled = true;
				goto TEST_NEXT_OP;
			}
		}
		
	TEST_NEXT_OP: ;		
	}	
}


void Task::build_dependencies(){
	assert( transGraphNum!= 0 );
	cout<<"Building Dependencies...  ";
	unsigned int i,j,k,m,g;
	TransitionGraph *tg = NULL;
	
	//Initialize dependencyList;
	int checkList[MAX_DTG_NUM];
	memset(checkList, 0, transGraphNum * sizeof(int) );
	for(g=0;g<transGraphNum;g++){
		tg = &(transGraph[g]);
		memset(checkList, 0, (sizeof(int))* transGraphNum);
			
		for(i=0;i<tg->size;i++)
		for(j=0;j<tg->size;j++){
			
			TransitionEdge& edge = tg->edge[i][j];
			if( edge.num_action == 0 ) continue;
			
			for(k=0;k<edge.num_action;k++){
				int action = edge.action[k];
				for(m=0;m<gop[action].num_P;m++){
					unsigned graphId = gft[gop[action].P[m]].graph;
					if( graphId == UNDEFINED ){
						print_op_name(action);cout<<endl;
						printf("    ");print_ft_name(gop[action].P[m]);cout<<endl;
					} 
					assert( graphId != UNDEFINED );
					if( graphId == g ) continue;
					
					checkList[ gft[gop[action].P[m]].graph ] = 1;
				}
			}
		}
		
		for(i=0;i<transGraphNum;i++)
			if( checkList[i]==1 ) tg->dependencyList.push_back( i );
	}
	
	
	//Initialize DependencyMatrix (via downward);
	causalMatrix = new bool*[transGraphNum];
	for(g=0;g<transGraphNum;g++){
		tg = &(transGraph[g]);
		causalMatrix[g] = new bool[transGraphNum];
		memset( causalMatrix[g], false , transGraphNum * sizeof(bool) );
	}
	read_causal_graphs( causalMatrix, var_mapping, transGraphNum );
	
	//Compute graph dependency degree
	for(i=0;i<transGraphNum;i++){
		for(j=0;j<transGraphNum;j++){
			if( causalMatrix[i][j] )
				transGraph[i].dependToDegree++;
			if( causalMatrix[j][i] )
				transGraph[i].dependByDegree++;			
		}
	}
	
	//Transition-Processed casual Matrix;
	transitCausalMatrix = new bool*[transGraphNum];
	for(g=0;g<transGraphNum;g++){
		tg = &(transGraph[g]);
		transitCausalMatrix[g] = new bool[transGraphNum];
		for(i=0;i<transGraphNum;i++)
			transitCausalMatrix[g][i] = causalMatrix[g][i];
	}	
	for(i=0;i<transGraphNum;i++)
		for(j=0;j<transGraphNum;j++)
			for(k=0;k<transGraphNum;k++){
				if( transitCausalMatrix[i][j] && transitCausalMatrix[j][k] )
					transitCausalMatrix[i][k] = true;
			}
	
	cout<<"Done."<<endl;
	
	//Debug Info;
	if( solverContext->display_info == 110 ){
		for(i=0;i<transGraphNum;i++){
			cout<<"Graph#"<<i<<" depends on:";
			for(j=0;j<transGraphNum;j++) 
				if( causalMatrix[i][j] )  cout<<j<<" ";
			cout<<endl;
		}
		exit(0);
	}
}
