/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/




#include<stdio.h>
#include<vector>
#include<iostream>
#include<fstream>

#include "state.h"
#include "plan.h"
#include "task.h"

#include "transition_graph.h"

using namespace std;

static Task&task = Task::getInstance();


/*****************************************
 *				TransitionVertex;;;;;;;
 *****************************************/

void TransitionVertex::set_value(vector<int>& list){//Multiple Element
	for(unsigned i=0;i<list.size();i++)
		this->elems[i] = list[i];
	dimension = list.size();
}

void TransitionVertex::set_value(int elem){//Single Element;
	this->elems[0] = elem;
	dimension = 1;
}


bool TransitionVertex::has_fact(unsigned fact){
	for(unsigned i=0;i<dimension;i++)
		if( elems[i] == fact ) 
			return true;
	return false;
}


unsigned& TransitionVertex::operator[](unsigned offset){
	assert( offset < dimension);
	return elems[offset];
}


void TransitionVertex::print(){
	Task& task = Task::getInstance();
	printf("{");
	for(unsigned i=0;i<dimension;i++){
		unsigned ft = elems[i];
		if( ft <task.gnum_ft )
			task.gft[elems[i]].print_name();
		else{
			//assert( ft - task.gnum_ft <= task.graphGroup.size() );
			cout<<"(NONE-"<< graph->id <<")";
		}
	}
	printf("}");
}

bool TransitionVertex::isNone(){
	
	Task& task = Task::getInstance();
	if( this->elems[0] >= task.gnum_ft )
		return true;
	else return false;
}



void TransitionVertex::println(){
	print();
	cout<<endl;
}


/*****************************************
 *				TransitionEdge;
 *****************************************/
void TransitionEdge::print(){
	cout<<"[from";
	graph->print_vertex(fromIndex);
	cout<<" to";
	graph->print_vertex(toIndex);	
}

void TransitionEdge::println(){
	print();
	cout<<endl;
}


void TransitionEdge::println(const char* str){
	cout<<str;
	println();
}

/*****************************************
 *				TransitionGraph;
 *****************************************/


TransitionGraph::TransitionGraph(){
	id 			= UNDEFINED;
	isMono 		= false;
	isAxiom 	= false;
	invariantId = UNDEFINED;
	dimension 	= 0;
	size 		= 0;
	vertex 		= NULL;
	edge 		= NULL;
	path 		= NULL;
	distance 	= NULL;
	ptr_task 	= NULL;
	hasNegateAll= false;
	
	dependToDegree 	= 0;
	dependByDegree 	= 0;
	dlevel		   	= UNDEFINED;
}

TransitionGraph::~TransitionGraph(){																//To-do:it's important to free them correctly.
	unsigned i;
	delete[] vertex;

	if(edge!=NULL){
		for(i=0;i<size;i++)
			delete[] edge[i];
		delete[] edge;
	}
	if( path!= NULL ){
		for(i=0;i<size;i++)
			delete[] path[i];
		delete[] path;
	}
	
	if( distance!=NULL ){
		for(i=0;i<size;i++)
			delete[] distance[i];
		delete[] distance;
	}
}


bool TransitionGraph::has_fact(unsigned fact){
	
	for(unsigned i=0;i<size;i++)
		if( vertex[i].has_fact(fact))
			return true;
	return false;
}

bool TransitionGraph::has_oper(unsigned oper){
	
	return false;
}


void TransitionGraph::build_vertexes(vector<unsigned>& facts){
	
	assert( this->id != UNDEFINED ); 
	unsigned i;

    vertex = new TransitionVertex[facts.size()];
    this->size = facts.size();
    dimension = 1;

    for( i=0 ; i<facts.size() ; i++ ){
		vertex[i].graph = this;
		if( facts[i] != UNDEFINED )
			vertex[i].set_value( facts[i] );
		else{
			vertex[i].set_value( task.gnum_ft + this->id );
			this->hasNegateAll = true;
		}
    }
}


int TransitionGraph::depend_degree(){
	return dependToDegree - dependByDegree;
}

void TransitionGraph::build_edges(){
	unsigned i,j,k,l,m;	
	Task& task = Task::getInstance();
    vector<unsigned> actionList;
    
	//Initialize Edges;
    edge = new TransitionEdge*[size];
    for(i=0;i<size;i++){
    	edge[i] = new TransitionEdge[size];
    	for(j=0;j<size;j++)
    		edge[i][j].graph = this;
    }
    
	//Build edges; 
    for( i=0;i<size;i++)
    for( j=0;j<size;j++){
		if(i==j) continue;

		assert( vertex[i][0] <  task.gnum_ft  || vertex[j][0] <  task.gnum_ft);
		actionList.clear();
		
    	if( vertex[i][0] >= task.gnum_ft  ){
    		
    		//Special case: NONE-(X) node is pred vertex;
    		for(k=0;k<task.gft[vertex[j][0]].num_A;k++){
    			
    			//Condition: Any fact in this graph cannot exist in the actions' $del-effect$
    			int actionId = task.gft[ vertex[j][0]].A[k];
    			for(l=0;l<task.gop[actionId].num_D;l++)
    				if( has_fact( task.gop[actionId].D[l] ) ) 
    					break;
    			if( l < task.gop[actionId].num_D && task.gop[actionId].num_D !=0)
    				continue;
    			
    			//Condition:  the succ fact/vertex (vertex[j][0]) shouldn't be actionId's precondition;
    			for(l=0;l<task.gop[actionId].num_P;l++)
    				if( vertex[j][0] == task.gop[actionId].P[l] ) break;
    			if( l < task.gop[actionId].num_P){
    				continue;
    			} 
    			
    			// Condition '==0' is important 
    			actionList.push_back(actionId);
    		}

    	}else if( vertex[j][0] >=  task.gnum_ft  ){
    		
    		//Special Case; NONE-(X) node is succ vertex;    		
    		// We may need to add an constraint here, which restricts the fact must be in both D and P

    		for(k=0;k<task.gft[vertex[i][0]].num_D;k++){
    			//Condition: Any fact in this graph cannot exist in the actions' $add-effect$    			
    			int actionId = task.gft[vertex[i][0]].D[k];

    			for(l=0;l<task.gop[actionId].num_A;l++){
    				if( has_fact( task.gop[actionId].A[l] ) ) 
    					break;
    			}
    			
    			if( task.gop[actionId].num_A == 0 || l == task.gop[actionId].num_A) 
    				actionList.push_back(actionId);
    		}
    		
    	}else{
    		
    		//NORMAL CASE:
			for(k=0;k<task.gft[vertex[i][0]].num_D;k++){
				int oper = 	task.gft[vertex[i][0]].D[k];
				for( l=0;l<task.gop[oper].num_A;l++)
					if( vertex[j][0] == task.gop[oper].A[l] ) 
						break;
				if( l < task.gop[oper].num_A )
			    	actionList.push_back(oper);
			}
    	}
    	
	    unsigned& num_action = edge[i][j].num_action = 0;	
		if ( actionList.size()!=0 ){
			edge[i][j].fromIndex = i;
			edge[i][j].toIndex = j;
		    edge[i][j].action = new unsigned[actionList.size()];
		    for( k=0;k<actionList.size();k++ ){
		    	if( !task.gop[actionList[k]].isDisabled )
					edge[i][j].action[num_action++] = actionList[k];
		    }
		}
		
		if( num_action > 0 ){
		    //Compute shared preconditions;
		    unsigned firstOp = edge[i][j].action[0];

		    for(k=0;k<task.gop[firstOp].num_P;k++){
		    	unsigned checkingP = task.gop[firstOp].P[k];
		    	
		    	for(l=1;l<edge[i][j].num_action;l++){
		    		int op = edge[i][j].action[l];
		    		for( m=0; m<task.gop[op].num_P; m++ )
		    			if( task.gop[op].P[m] == checkingP ) break;
		    		
		    		if( m==task.gop[op].num_P ) break;
		    	}
		    	
		    	if( l == edge[i][j].num_action && checkingP!= i)
		    		edge[i][j].SP[edge[i][j].num_SP++] = checkingP;
		    }
		}
    }
}


void TransitionGraph::build_inner_path( ){
	
	unsigned fromIndex;
	unsigned i,j,k,oper,ft;
	for(i=0;i<size;i++)
	for(j=0;j<size;j++){
		TransitionEdge& e = edge[i][j];
		if( e.num_action > 1 ) return;
		if( e.num_action == 0 ) continue;
		
		oper = e.action[0];
		for(k=0;k<task.gop[oper].num_P;k++){
			ft = task.gop[oper].P[k];
			if ( ft >=0 && ft<task.gnum_ft && task.gft[ft].graph != id ) 
				return;
		}
	}
	
	path = new Plan*[size];
	for(i=0;i<size;i++){
		path[i] = new Plan[size];
		for(j=0;j<size;j++)
			path[i][j].set_infinity();
	}
	
	for( fromIndex = 0 ; fromIndex< size; fromIndex ++ ){
	    int visited[MAX_DTG_SIZE];

	    vector<unsigned> expand_list;
	    expand_list.push_back(fromIndex);
	    
	    Plan *wpath = path[fromIndex];
	    wpath[fromIndex].set_empty();
		vector<unsigned> tmp_list;
	    
		//Plan singleAction;
		//singleAction.length = 1;
		
	    for( i=0 ; i<size; i++ )
			visited[i] = 0;
	
	    do{
			for( i=0 ; i< expand_list.size() ; i++)
			    visited[ expand_list[i] ] = 1;
	
			tmp_list.clear();
			
			for( i=0 ; i<expand_list.size(); i++)
			    for( j=0 ; j<size; j++ ){
			    	
			    	TransitionEdge& next_edge = edge[expand_list[i]][j]; 
				    if ( next_edge.num_action!=0  && !visited[j] ) {
				    	
						wpath[j].copy(wpath[expand_list[i]]);

						assert( next_edge.num_action == 1);
						wpath[j].add_oper( next_edge.action[0] );
						
						tmp_list.push_back(j);
				    }
			    }
	
		    expand_list.clear();
		    for( i=0;i<tmp_list.size();i++ )
		    	expand_list.push_back( tmp_list[i] );
		    	
	    }while( expand_list.size() );
	}
}


void TransitionGraph::print(){
	unsigned i,j,l,k,edgeCount=0;
	Task& task = Task::getInstance();
	
	for(i=0;i<size;i++)
	for(j=0;j<size;j++){
		TransitionEdge& e = this->edge[i][j];
		if( e.num_action!=0 )
			edgeCount++;
	}

	printf("Graph#%d: [S%d D%d I%u ]; ", id, size,dlevel,type_identifier);
	for(i=0;i<size;i++)
	    vertex[i].print();
	printf("\n");
	for(j=0;j<size;j++)
    for(k=0;k<size;k++){
    	if( edge == NULL ) break;
    	
	    TransitionEdge& e = edge[j][k];
	    
	    if( e.num_action>0 ){

			vertex[j].print();
			printf(" -->> ");
			vertex[k].print();
			printf("Action_NUM:(%d)", e.num_action);
			
			printf("\n");fflush(stdout);
			
			for( l=0 ; l < e.num_action ; l++ ) {
			    printf("   [%d",e.action[l]);
			    task.gop[e.action[l]].print_name();
			    printf("[PRE:%d]\n",task.gop[e.action[l]].num_P);
			}
	    }
    }
   
   	//Print out inner-paths
   	if( this->path!= NULL){
		for( i=0;i<size;i++ )
		for( j=0;j<size;j++ ){
			if( i == j ) continue;
			char word[256];
			sprintf(word,"P from %d to %d",i,j);
			if( !path[i][j].is_infinity() ) path[i][j].println(word);
		}
   	}
    printf("\n");
}

/*
void TransitionGraph::print_vertexes(){
	
	for(unsigned i=0;i<size;i++)
		cout<< vertex[i][0] << ";";
	printf("\n");
} */


void TransitionGraph::print_vertex_only(){
	unsigned i;

	printf("Graph#%d: [S%d D%d I%u ]; ", id, size,dlevel,type_identifier);
	for(i=0;i<size;i++) vertex[i].print();
	printf("\n");
}


void TransitionGraph::calc_distances(){
	unsigned i,j,k; 
	assert( size != 0);
	
	distance = new unsigned*[size];
	for(i=0;i<size;i++){
		distance[i] = new unsigned[size];
		for(j=0;j<size;j++)
			distance[i][j] = INF;
	}
	
	for(i=0;i<size;i++)
	for(j=0;j<size;j++)
		if( i== j )
			distance[i][j] = 0;
		else if( edge[i][j].num_action!=0)
			distance[i][j]=1;
		
	for(k=0;k<size;k++){
		for(i=0;i<size;i++)
		for(j=0;j<size;j++){
			unsigned a = distance[i][k] + distance[k][j]; 
			if(  a < distance[i][j])
				distance[i][j] = a;
		}
	}
}

///	Compute a unique idenfier for the graph, by counting the predicates;
unsigned TransitionGraph::compute_identifier(){
	unsigned i,result = UNDEFINED;
	
	bool* flag = (bool*) malloc( sizeof(bool) * task.predicates.size());
	memset( flag, false, sizeof(bool) * task.predicates.size());
	
	for(i=0;i<size;i++){
		if( vertex[i][0] >= task.gnum_ft ) continue;
		assert( task.gft[vertex[i][0]].predicate != UNDEFINED );
		flag[ task.gft[vertex[i][0]].predicate ] = true;
	}
	
	result = 0 ;
	for(i=0;i<task.predicates.size();i++)
		if( flag[i] )
			result = result * 10 + i * i ;
	
	if( result == UNDEFINED )
		result = this->id;
	type_identifier = result;
	
	free( flag );
	return result;
}


/**************************************************************************************/
/*********************************  LOCAL Helper actions  *****************************/

unsigned w[1000];
unsigned w_len;
unsigned w_flag[1000];
unsigned min_path_len = 2000;

void TransitionGraph::dtg_path_depth_search(
							unsigned currentIndex, 
							unsigned goalIndex, 
							unsigned ** pathList, 
							unsigned *len, 
							unsigned *pathNum){

	unsigned i;

	w[w_len++] = currentIndex;
	w_flag[currentIndex] = 1;
	
	if( currentIndex == goalIndex  ){
		
		if( w_len < min_path_len ){  //This routine could be improved by another way of path retrieving;
			unsigned num = *pathNum; 
			pathList[num] = new unsigned[w_len];
			len[num] = w_len;
			for(i=0;i<w_len;i++)
				pathList[num][i] = w[i];
			(*pathNum)++;
			min_path_len = w_len;
		}
	}else{
		for(i=0;i<size;i++){
			TransitionEdge& e = this->edge[currentIndex][i];

			if( e.num_action!=0 && !w_flag[i] ){
				dtg_path_depth_search( i , goalIndex, pathList, len , pathNum );			
			}
		}
	}
	
	w_flag[currentIndex] = 0;	
	w_len--;
}

unsigned TransitionGraph::retrieve_all_path(
							unsigned fromIndex, 
							unsigned toIndex, 
							unsigned** pathList, 
							unsigned* len){
	
	w_len = 0;
	min_path_len = 2000;
	memset( w, 0, sizeof(unsigned) * 1000 );
	memset( w_flag , 0, sizeof(unsigned)*1000 );
	
	unsigned num = 0;
	dtg_path_depth_search( fromIndex , toIndex, pathList, len,  &num);
	
	return num;
}

void TransitionGraph::print_vertex(int index){
	vertex[index].print();
}


void TransitionGraph::print_edge(int from, int to){
	edge[from][to].print();
}


void TransitionGraph::print_with_full_detail(){
	unsigned i,j,l,k,edgeCount=0;
	Task& task = Task::getInstance();
	
	for(i=0;i<size;i++)
	for(j=0;j<size;j++){
		TransitionEdge& e = this->edge[i][j];
		if( e.num_action!=0 )
			edgeCount++;
	}

	printf("Graph#%d:[%d]; ", id, size);
	/*
	if(graph->actor==NULL) printf("[Object: n_ptr]\n");
	    else printf("[actor:%s]" , graph->actor);
	    
	printf("[Edge Count:%d][invar%d]\n",edgeCount,graph->invariantId);
	printf("Group elements:");
	*/
	
	for(i=0;i<size;i++)
	    vertex[i].print();
	printf("\n");

	
	for(j=0;j<size;j++)
    for(k=0;k<size;k++){
    	if( edge == NULL ) break;
    	
	    TransitionEdge& e = edge[j][k];
	    
	    if( e.num_action>0 ){

			vertex[j].print();
			printf(" -->> ");
			vertex[k].print();
			printf("Action_NUM:(%d)", e.num_action);
			
//			printf("SP:(");
//			for(i=0;i<edge->num_SP;i++)
//				print_ft_name(edge->SP[i]),printf(";");
//			printf(")");
//			printf("FP:(");
//			for(i=0;i<edge->num_FP;i++)
//				print_ft_name(edge->FP[i]),printf(";");
//			printf(")");
			
			printf("\n");fflush(stdout);
			
			for( l=0 ; l < e.num_action ; l++ ) {
			    printf("   [");
			    task.gop[e.action[l]].print_name();
			    printf("[PRE:%d]",task.gop[e.action[l]].num_P);
	
			    printf("] {");

//			    Action * action = gop_conn[edge->action[l]];
//			    for(m=0;m<dis_gop_conn[edge->action[l].num_P;m++){
//				    int condition = dis_gop_conn[edge->action[l].P[m];
//				    if(condition==graph->element[j]) continue;
//				    dis_print_ft_name( condition );
//				    printf(";");
//				}
			    
			    printf("}\n");
			}
			
	    }
    }
   
   	//Print out inner-paths
   	if( this->path!= NULL){
		for( i=0;i<size;i++ )
		for( j=0;j<size;j++ ){
			if( i == j ) continue;
			char word[256];
			sprintf(word,"P from %d to %d",i,j);
			path[i][j].println(word);
		}
   	}
    printf("\n");
}
