/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/



#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>

#include "state.h"
#include "plan.h"
#include "utils.h"

using namespace std;

float getVmRSS(){
	int pid;
	char buffer[1024];
	FILE* fp;
	pid = getpid();
	sprintf( buffer, "/proc/%i/status", pid );
	
	fp = fopen( buffer, "r" );
	
	if(fp==NULL) printf("ERROR while reading VmRSS!!!!\n"),exit(0);
	
	char columnHead[50] = "VmRSS";
	while(fgets(buffer,1024,fp)!=NULL){
		if(strncmp(buffer,columnHead,strlen(columnHead))==0){
			unsigned i, start=0,end=0 ;
			for(i=0;i<strlen(buffer);i++)
				if(isdigit(buffer[i])) {
					start = i;
					break;
				}
			
			do{
				if(!isdigit(buffer[i])){
					end = i;
					break;
				}
			}while(++i<strlen(buffer));
			
			char ns[50];
			for(i=0;i<end-start;i++) ns[i]= buffer[start+i];
			ns[i++] = '\0';
			int num = atoi(ns);
			fclose(fp);
			return (float)num/1024;
		}
	}
	
	//FAIL to find one.
	printf("Something is wrong with the VmRSS reading.!!!!!!!!!!!!!\n");
	return -1;	
}


double get_global_time(){
	struct rusage us;
	getrusage(RUSAGE_SELF, &us);
	double res = (double)us.ru_utime.tv_sec + (double)us.ru_utime.tv_usec/1000000;
	
	getrusage(RUSAGE_CHILDREN, &us);
	res += ( us.ru_utime.tv_sec +  (double)us.ru_utime.tv_usec/1000000 );
	return res;
}


long get_global_time_int(){
	struct rusage us;
	getrusage(RUSAGE_SELF, &us);
	return us.ru_utime.tv_sec;	
}


void Error(const char* str){
	printf("%s",str);
	exit(1);
}


void clean_up_array(unsigned* array, unsigned len, vector<unsigned>& to_remove){
	
}


int bsearch_vector(vector<unsigned>& vec, unsigned value){
	int low = 0;
    int high = vec.size() - 1;
    while (low < high) {
        int mid = (low + high)/2;
        if ( vec[mid] <value )
            low = mid + 1;
        else
            high = mid; 
    }
    
    if ( low <  (int)vec.size() &&  vec[low] == value ){
		return  low;		//assert(false);
    }else{
    	return -1;
    }
}



unsigned DJBHash(unsigned* str, unsigned len){  
   unsigned hash = 5381;
   unsigned i    = 0;

   for(i = 0; i < len; str++, i++)
      hash = ((hash << 4) + hash) + (*str);

   return hash;
}


unsigned DJBHash(vector<unsigned>& vec){
   unsigned hash = 5381;
   unsigned i    = 0;

   for(i = 0; i < vec.size(); i++)
      hash = ((hash << 4) + hash) + vec[i];

   return hash;
}


//End of File

